from enum import Enum


class OrganizationInvitationStatus(str, Enum):
    ACCEPTED = "ACCEPTED"
    PENDING = "PENDING"
    REJECTED = "REJECTED"

    def __str__(self) -> str:
        return str(self.value)



import sklearn.pipeline
import copy
from jaqpotpy.models.base_classes import Model
from jaqpotpy.doa.doa import DOA
from typing import Any, Union, Dict, Optional
from jaqpotpy.datasets.molecular_datasets import JaqpotpyDataset
from jaqpotpy.models import Evaluator, Preprocess
from jaqpotpy.api.get_installed_libraries import get_installed_libraries
from jaqpotpy.api.types.models import FeatureType
from jaqpotpy.api.types.models import ModelType
import sklearn
from jaqpotpy.cfg import config
import jaqpotpy
from skl2onnx import convert_sklearn
from skl2onnx.common.data_types import FloatTensorType
from onnxruntime import InferenceSession
from onnxruntime import InferenceSession
import numpy as np

class SklearnModel(Model):

    def __init__(self, dataset: JaqpotpyDataset, doa: DOA, model: Any,
                  preprocessor: Preprocess = None,  evaluator: Evaluator = None):
        self.x_cols = dataset.x_cols
        self.y_cols = dataset.y_cols
        self.dataset = dataset
        self.featurizer = dataset.featurizer
        self.model = model
        self.pipeline = None
        self.pipeline = None
        self.trained_model = None
        self.doa = doa
        self.trained_doa = None
        self.evaluator = evaluator
        self.preprocess = preprocessor
        self.preprocessing_y = None
        self.libraries = None
        self.version = [sklearn.__version__]
        self.jaqpotpy_version = jaqpotpy.__version__
        self.jaqpotpy_docker = config.jaqpotpy_docker
        self.task = self.dataset.task
        self.onnx_model = None
        self.onnx_opset = None
        self.type = ModelType('SKLEARN')
        self.independentFeatures = None
        self.dependentFeatures = None

    def __dtypes_to_jaqpotypes__(self):
        for feature in self.independentFeatures + self.dependentFeatures:
            if feature['featureType'] in ['SMILES']:
                feature['featureType'] = FeatureType.SMILES
            elif feature['featureType'] in ['int', 'int64']:
                feature['featureType'] = FeatureType.INTEGER
            elif feature['featureType'] in ['float', 'float64']:
                feature['featureType'] = FeatureType.FLOAT
            elif feature['featureType'] in ['string, object']:
                feature['featureType'] = FeatureType.STRING

    def fit(self, onnx_options: Optional[Dict] = None):
        self.libraries = get_installed_libraries()

        if self.dataset.y is None:
            raise TypeError("dataset.y is None. Please provide a target variable for the model.")
        #Get X and y from dataset
        X = self.dataset.__get_X__()
        y = self.dataset.__get_Y__()

        if self.doa:
            self.trained_doa = self.doa.fit(X=X)
        
        if len(self.dataset.y_cols) == 1:
            y = y.to_numpy().ravel()

        #if preprocessing was applied to either X,y or both
        if self.preprocess is not None:
            self.pipeline = sklearn.pipeline.Pipeline(steps=[])
            self.pipeline = sklearn.pipeline.Pipeline(steps=[])
            # Apply preprocessing on design matrix X
            pre_keys = self.preprocess.classes.keys()
            if len(pre_keys) > 0:
                for preprocessor in pre_keys:
                    self.pipeline.steps.append((preprocessor, self.preprocess.classes.get(preprocessor)))
            self.pipeline.steps.append(('model', self.model))

            # Apply preprocessing of response vector y
            pre_y_keys = self.preprocess.classes_y.keys()

            if len(pre_y_keys) > 0:
                if self.task == "classification":
                    raise ValueError("Target labels cannot be preprocessed for classification tasks. Remove any assigned preprocessing for y.")
                else:
                    preprocess_names_y = []
                    preprocess_classes_y = []
                    y_scaled = self.dataset.__get_Y__()
                    for pre_y_key in pre_y_keys:
                        pre_y_function = self.preprocess.classes_y.get(pre_y_key)
                        y_scaled = pre_y_function.fit_transform(y_scaled)
                        if len(self.dataset.y_cols) == 1:
                            y_scaled = y_scaled.ravel()
                        self.preprocess.register_fitted_class_y(pre_y_key, pre_y_function)
                        preprocess_names_y.append(pre_y_key)
                        preprocess_classes_y.append(pre_y_function)
                    self.preprocessing_y = preprocess_classes_y

                    self.trained_model = self.pipeline.fit(X.to_numpy(), y_scaled)
            else:
                self.trained_model = self.pipeline.fit(X.to_numpy(), y)#.to_numpy().ravel())   
        #case where no preprocessing was provided
        else:
            self.trained_model = self.model.fit(X.to_numpy(), y)#.to_numpy().ravel())
        
        if self.dataset.smiles_cols:
            self.independentFeatures = list({"key": self.dataset.smiles_cols[smiles_col_i], "name": self.dataset.smiles_cols[smiles_col_i], "featureType": "SMILES"} for smiles_col_i in range(len(self.dataset.smiles_cols)))
        else:
            self.independentFeatures = list()
        if self.dataset.x_cols:
            self.independentFeatures += list({"key": feature, "name": feature, "featureType": X[feature].dtype} for feature in self.dataset.x_cols)
        self.dependentFeatures = list({"key": feature, "name": feature, "featureType": self.dataset.__get_Y__()[feature].dtype} for feature in self.dataset.y_cols)
        self.__dtypes_to_jaqpotypes__()

        name = self.model.__class__.__name__ + "_ONNX"
        self.onnx_model = convert_sklearn(self.trained_model, initial_types=[('float_input', FloatTensorType([None, X.to_numpy().shape[1]]))], name=name,
                                          options=onnx_options)
        self.onnx_opset = self.onnx_model.opset_import[0].version
        
        if self.evaluator:
            self.__eval__()
        return self

    def predict(self,  dataset: JaqpotpyDataset):
        if not isinstance(dataset, JaqpotpyDataset):
            raise TypeError("Expected dataset to be of type JaqpotpyDataset")
        sklearn_prediction = self.trained_model.predict(dataset.X.to_numpy().astype(np.float32))
        if self.preprocess is not None:
            if self.preprocessing_y:
                for f in self.preprocessing_y:
                    if len(self.y_cols) == 1:
                        sklearn_prediction = f.inverse_transform(sklearn_prediction.reshape(1, -1)).flatten()
                    else:
                        sklearn_prediction = f.inverse_transform(sklearn_prediction)
        return sklearn_prediction
    
    def predict_proba(self, dataset: JaqpotpyDataset):
        if not isinstance(dataset, JaqpotpyDataset):
            raise TypeError("Expected dataset to be of type JaqpotpyDataset")
        if self.task == "regression":
            raise ValueError("predict_proba is available only for classification tasks")
        sklearn_probs = self.trained_model.predict_proba(dataset.X.to_numpy())
        sklearn_probs_list = [max(sklearn_probs[instance]) for instance in range(len(sklearn_probs))]
        return sklearn_probs_list
    
    def predict_onnx(self,  dataset: JaqpotpyDataset):
        if not isinstance(dataset, JaqpotpyDataset):
            raise TypeError("Expected dataset to be of type JaqpotpyDataset")
        sess = InferenceSession(self.onnx_model.SerializeToString())
        onnx_prediction = sess.run(None, {"float_input": dataset.X.to_numpy().astype(np.float32)})
        if len(self.y_cols) == 1:
            onnx_prediction[0] = onnx_prediction[0].reshape(-1, 1)
        if self.preprocess is not None:
            if self.preprocessing_y:
                for f in self.preprocessing_y:
                    onnx_prediction[0] = f.inverse_transform(onnx_prediction[0])
        if len(self.y_cols) == 1:
            return onnx_prediction[0].flatten()
        return onnx_prediction[0]
    
    def predict_proba_onnx(self, dataset: JaqpotpyDataset):
        if not isinstance(dataset, JaqpotpyDataset):
            raise TypeError("Expected dataset to be of type JaqpotpyDataset")
        if self.task == "regression":
            raise ValueError("predict_onnx_proba is available only for classification tasks")
        sess = InferenceSession(self.onnx_model.SerializeToString())
        onnx_probs = sess.run(None, {"float_input": dataset.X.to_numpy().astype(np.float32)})
        onnx_probs_list = [max(onnx_probs[1][instance].values()) for  instance in range(len(onnx_probs[1]))]
        return onnx_probs_list

    def __eval__(self):
        if self.evaluator.dataset.df is not None:
            pass
        else:
            self.evaluator.dataset.create()
        X = self.evaluator.dataset.__get_X__()
        if self.preprocess:
            pre_keys = self.preprocess.classes.keys()
            for pre_key in pre_keys:
                preprocess_func = self.preprocess.fitted_classes.get(pre_key)
                X = preprocess_func.transform(X)
        sess = InferenceSession(self.onnx_model.SerializeToString())
        sess = InferenceSession(self.onnx_model.SerializeToString())
        input_name = sess.get_inputs()[0].name
        X = np.array(X.astype(float).copy())
        preds = sess.run(None, {input_name: X.astype(np.float32)})
        preds = preds[0].flatten()
        # preds = self.trained_model.predict(X)
        preds_t = []
        for p in preds:
            try:
                if self.preprocessing_y:
                    for f in self.preprocessing_y:
                        p = f.inverse_transform(p.reshape(1, -1))
                        preds_t.append(p)
            except AttributeError as e:
                pass
            preds_t.append(p)
        eval_keys = self.evaluator.functions.keys()
        for eval_key in eval_keys:
            eval_function = self.evaluator.functions.get(eval_key)
            try:
                if self.preprocessing_y:
                    for f in self.preprocessing_y:
                        truth = f.inverse_transform(self.evaluator.dataset.__get_Y__())
                    print(eval_key + ": " + str(eval_function(truth, preds_t)))
            except AttributeError as e:
                print(eval_key + ": " + str(eval_function(self.evaluator.dataset.__get_Y__(), preds_t)))
                pass
            # print(eval_key + ": " + str(eval_function(self.evaluator.dataset.__get_Y__(), preds_t)))
            #print(eval_key + ": " + str(eval_function(self.evaluator.dataset.__get_Y__(), preds)))

    def copy(self):
        copied_model = copy.deepcopy(self)
        copied_model.dataset = None
        return copied_model
    
    def deploy_on_jaqpot(self, jaqpot, name, description, visibility):
        jaqpot.deploy_SklearnModel(model = self, name = name, description = description, visibility = visibility)



#from jaqpotpy.docking.pose_generation import PoseGenerator
#from jaqpotpy.docking.pose_generation import VinaPoseGenerator
#from jaqpotpy.docking.pose_generation import GninaPoseGenerator
#from jaqpotpy.docking.docking import Docker
#from jaqpotpy.docking.binding_pocket import ConvexHullPocketFinder


"""
Dataset abstract classes
"""
from abc import ABC, abstractmethod
import os
import pickle
from typing import Iterable, Optional
import pandas as pd

class BaseDataset(ABC):
    """
    Abstract class for datasets. This class defines the common interface and basic functionality 
    for dataset manipulation and handling.

    Attributes:
        _df (pd.DataFrame): The underlying DataFrame holding the dataset.
        x_cols (Optional[Iterable[str]]): The columns to be used as features.
        y_cols (Optional[Iterable[str]]): The columns to be used as labels.
        _task (str): The task type, either 'regression' or 'classification'.
        _dataset_name (str): The name of the dataset.
        _y (Iterable[str]): The labels of the dataset.
        _x (Iterable[str]): The features of the dataset.
    """

    def __init__(self, df: pd.DataFrame = None, path: Optional[str] = None,
                 y_cols: Iterable[str] = None,
                 x_cols: Optional[Iterable[str]] = None,
                 task: str = None) -> None:

        if df is None and path is None:
            raise TypeError("Either a DataFrame or a path to a file must be provided.")
        elif (df is not None) and (path is not None):
            raise TypeError("Either a DataFrame or a path to a file must be provided.")

        if df is not None:
            if not isinstance(df, pd.DataFrame):
                raise TypeError("Provided 'df' must be a pandas DataFrame.")
            else:
                self._df = df
                self.path = None
        elif path is not None:
            self.path = path
            extension = os.path.splitext(self.path)[1]
            if extension == '.csv':
                self._df = pd.read_csv(path)
            else:
                raise ValueError("The provided file is not a valid CSV file.")
            
        if not(isinstance(y_cols, str) or 
               (isinstance(y_cols, list) and all(isinstance(item, str) for item in y_cols)) or
               (y_cols is None)):
               raise TypeError("y_cols must be provided and should be either" 
                               "a string or a list of strings, or None") 
        
        if not(isinstance(x_cols, str) or 
              (isinstance(x_cols, list) and all(isinstance(item, str) for item in x_cols)) or
              (isinstance(x_cols, list) and len(x_cols) == 0) or
              (x_cols is None)):
               raise TypeError("x_cols should be either a string, an empty list"
                               "a list of strings, or None") 

        #Find the length of each provided column name vector and put everything in lists
        if isinstance(y_cols, str):
            self.y_cols = [y_cols]
            self.y_cols_len = 1
        elif isinstance(y_cols, list) :
            self.y_cols = y_cols
            self.y_cols_len = len(y_cols)
        elif y_cols is None:
            self.y_cols= []
            self.y_cols_len = 0

        if isinstance(x_cols, str):
            self.x_cols = [x_cols]
            self.x_cols_len = 1
        elif isinstance(x_cols, list) :
            self.x_cols = x_cols
            self.x_cols_len = len(x_cols)
        elif x_cols is None:
            self.x_cols= []
            self.x_cols_len = 0
        
        self.task = task
        self._dataset_name = None
        self._y = None
        self._x = None

    @property
    def df(self) -> pd.DataFrame:
        return self._df

    @df.setter
    def df(self, value: pd.DataFrame):
        if not isinstance(value, pd.DataFrame):
            raise ValueError("The value must be a pandas DataFrame.")
        self._df = value

    @df.deleter
    def df(self):
        del self._df

    @property
    def task(self):
        return self._task

    @task.setter
    def task(self, value):
        if value is None:
            raise ValueError("Task must be either 'regression' or 'classification'")
        elif value.lower() not in ['regression', 'classification']:
            raise ValueError("Task must be either 'regression' or 'classification'")
        self._task = value

    @property
    def dataset_name(self):
        return self._dataset_name

    @dataset_name.setter
    def dataset_name(self, value):
        self._dataset_name = value

    @property
    def X(self) -> Iterable[str]:
        return self._x

    @X.setter
    def X(self, value):
        self._x = value

    @property
    def y(self) -> Iterable[str]:
        return self._y

    @y.setter
    def y(self, value):
        self._y = value

    def save(self):
        if self._dataset_name:
            with open(self._dataset_name + ".jdata", 'wb') as f:
                pickle.dump(self, f)
        else:
            with open("jaqpot_dataset" + ".jdata", 'wb') as f:
                pickle.dump(self, f)

    @classmethod
    def load(cls, filename):
        with open(filename, 'rb') as f:
            return pickle.load(f)

    @abstractmethod
    def create(self):
        """
        Creates the dataset.
        """
        raise NotImplementedError
    
    @abstractmethod
    def __get_X__(self):
        """
        Returns the design matrix X.
        """
        raise NotImplementedError
    
    @abstractmethod
    def __get_Y__(self):
        """
        Returns the response Y.
        """
        raise NotImplementedError

    @abstractmethod
    def __repr__(self) -> str:
        """
        Returns a string representation of the dataset.
        """
        raise NotImplementedError

    @abstractmethod
    def __len__(self):
        """
        Returns the number of samples in the dataset.
        """
        raise NotImplementedError

    @abstractmethod
    def __get__(self, instance, owner):
        """
        Gets an attribute of the dataset.
        """
        raise NotImplementedError

    @abstractmethod
    def __getitem__(self, idx):
        """
        Gets a sample by index.
        """
        raise NotImplementedError


class MaterialDataset(BaseDataset):
    def __init__(self, df: pd.DataFrame = None, path:  Optional[str] = None,
                 y_cols: Iterable[str] = None,
                 x_cols: Optional[Iterable[str]] =None, materials_col=None,
                 materials=None) -> None:
        super().__init__(df = df, path = path, y_cols = y_cols, x_cols = x_cols)
        self.materials = materials
        self._materials_strings = None
        self.materials_col = materials_col

    @property
    def materials_strings(self) -> Iterable[str]:
        return self._materials_strings

    @materials_strings.setter
    def materials_strings(self, value):
        self._materials_strings = value


class ImageDataset(BaseDataset):
    def __init__(self,  df: pd.DataFrame = None, path:  Optional[str] = None,
                  y_cols: Iterable[str] = None,
                 x_cols: Optional[Iterable[str]] =None) -> None:
        super().__init__(df = df, path = path, y_cols = y_cols, x_cols = x_cols)


if __name__ == '__main__':
    ...



from enum import Enum


class FeatureType(str, Enum):
    CATEGORICAL = "CATEGORICAL"
    FLOAT = "FLOAT"
    INTEGER = "INTEGER"
    SMILES = "SMILES"
    STRING = "STRING"
    TEXT = "TEXT"

    def __str__(self) -> str:
        return str(self.value)



import pickle
from base64 import b64encode

def model_to_b64encoding(actualModel):
    p_mod = pickle.dumps(actualModel)
    raw = b64encode(p_mod)
    model_b64encoded = raw.decode('utf-8')
    return model_b64encoded


from typing import Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="User")


@_attrs_define
class User:
    """
    Attributes:
        id (str):
        name (Union[Unset, str]):
        email_verified (Union[Unset, bool]):
    """

    id: str
    name: Union[Unset, str] = UNSET
    email_verified: Union[Unset, bool] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = self.id

        name = self.name

        email_verified = self.email_verified

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
            }
        )
        if name is not UNSET:
            field_dict["name"] = name
        if email_verified is not UNSET:
            field_dict["emailVerified"] = email_verified

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        id = d.pop("id")

        name = d.pop("name", UNSET)

        email_verified = d.pop("emailVerified", UNSET)

        user = cls(
            id=id,
            name=name,
            email_verified=email_verified,
        )

        user.additional_properties = d
        return user

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.feature_feature_dependency import FeatureFeatureDependency
from ..models.feature_type import FeatureType
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.feature_meta import FeatureMeta


T = TypeVar("T", bound="Feature")


@_attrs_define
class Feature:
    """
    Attributes:
        key (str): A key that must start with a letter, followed by any combination of letters, digits, hyphens, or
            underscores. For example, 'abc123', 'abc-test', or 'Abc_test'. It cannot start with a digit. Example: feature-
            key.
        name (str): A name for the feature that will appear on top of the form field Example: A feature name.
        feature_type (FeatureType):  Example: FLOAT.
        id (Union[Unset, int]):  Example: 1.
        meta (Union[Unset, FeatureMeta]): A JSON object containing meta information.
        description (Union[Unset, str]):
        feature_dependency (Union[Unset, FeatureFeatureDependency]):  Example: DEPENDENT.
        visible (Union[Unset, bool]):  Example: True.
        possible_values (Union[Unset, List[str]]):
        created_at (Union[Unset, datetime.datetime]): The date and time when the feature was created. Example:
            2023-01-01T12:00:00Z.
        updated_at (Union[Unset, str]): The date and time when the feature was last updated. Example:
            2023-01-01T12:00:00Z.
    """

    key: str
    name: str
    feature_type: FeatureType
    id: Union[Unset, int] = UNSET
    meta: Union[Unset, "FeatureMeta"] = UNSET
    description: Union[Unset, str] = UNSET
    feature_dependency: Union[Unset, FeatureFeatureDependency] = UNSET
    visible: Union[Unset, bool] = UNSET
    possible_values: Union[Unset, List[str]] = UNSET
    created_at: Union[Unset, datetime.datetime] = UNSET
    updated_at: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        key = self.key

        name = self.name

        feature_type = self.feature_type.value

        id = self.id

        meta: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.meta, Unset):
            meta = self.meta.to_dict()

        description = self.description

        feature_dependency: Union[Unset, str] = UNSET
        if not isinstance(self.feature_dependency, Unset):
            feature_dependency = self.feature_dependency.value

        visible = self.visible

        possible_values: Union[Unset, List[str]] = UNSET
        if not isinstance(self.possible_values, Unset):
            possible_values = self.possible_values

        created_at: Union[Unset, str] = UNSET
        if not isinstance(self.created_at, Unset):
            created_at = self.created_at.isoformat()

        updated_at = self.updated_at

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "key": key,
                "name": name,
                "featureType": feature_type,
            }
        )
        if id is not UNSET:
            field_dict["id"] = id
        if meta is not UNSET:
            field_dict["meta"] = meta
        if description is not UNSET:
            field_dict["description"] = description
        if feature_dependency is not UNSET:
            field_dict["featureDependency"] = feature_dependency
        if visible is not UNSET:
            field_dict["visible"] = visible
        if possible_values is not UNSET:
            field_dict["possibleValues"] = possible_values
        if created_at is not UNSET:
            field_dict["createdAt"] = created_at
        if updated_at is not UNSET:
            field_dict["updatedAt"] = updated_at

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.feature_meta import FeatureMeta

        d = src_dict.copy()
        key = d.pop("key")

        name = d.pop("name")

        feature_type = FeatureType(d.pop("featureType"))

        id = d.pop("id", UNSET)

        _meta = d.pop("meta", UNSET)
        meta: Union[Unset, FeatureMeta]
        if isinstance(_meta, Unset):
            meta = UNSET
        else:
            meta = FeatureMeta.from_dict(_meta)

        description = d.pop("description", UNSET)

        _feature_dependency = d.pop("featureDependency", UNSET)
        feature_dependency: Union[Unset, FeatureFeatureDependency]
        if isinstance(_feature_dependency, Unset):
            feature_dependency = UNSET
        else:
            feature_dependency = FeatureFeatureDependency(_feature_dependency)

        visible = d.pop("visible", UNSET)

        possible_values = cast(List[str], d.pop("possibleValues", UNSET))

        _created_at = d.pop("createdAt", UNSET)
        created_at: Union[Unset, datetime.datetime]
        if isinstance(_created_at, Unset):
            created_at = UNSET
        else:
            created_at = isoparse(_created_at)

        updated_at = d.pop("updatedAt", UNSET)

        feature = cls(
            key=key,
            name=name,
            feature_type=feature_type,
            id=id,
            meta=meta,
            description=description,
            feature_dependency=feature_dependency,
            visible=visible,
            possible_values=possible_values,
            created_at=created_at,
            updated_at=updated_at,
        )

        feature.additional_properties = d
        return feature

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.dataset_entry_type import DatasetEntryType
from ..models.dataset_status import DatasetStatus
from ..models.dataset_type import DatasetType
from ..types import UNSET, Unset

T = TypeVar("T", bound="Dataset")


@_attrs_define
class Dataset:
    """
    Attributes:
        type (DatasetType):  Example: PREDICTION.
        entry_type (DatasetEntryType):  Example: ARRAY.
        input_ (List[Any]):
        id (Union[Unset, int]):  Example: 1.
        result (Union[Unset, List[Any]]):
        status (Union[Unset, DatasetStatus]):
        failure_reason (Union[Unset, str]):
        created_at (Union[Unset, str]):
        updated_at (Union[Unset, str]):
    """

    type: DatasetType
    entry_type: DatasetEntryType
    input_: List[Any]
    id: Union[Unset, int] = UNSET
    result: Union[Unset, List[Any]] = UNSET
    status: Union[Unset, DatasetStatus] = UNSET
    failure_reason: Union[Unset, str] = UNSET
    created_at: Union[Unset, str] = UNSET
    updated_at: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        type = self.type.value

        entry_type = self.entry_type.value

        input_ = self.input_

        id = self.id

        result: Union[Unset, List[Any]] = UNSET
        if not isinstance(self.result, Unset):
            result = self.result

        status: Union[Unset, str] = UNSET
        if not isinstance(self.status, Unset):
            status = self.status.value

        failure_reason = self.failure_reason

        created_at = self.created_at

        updated_at = self.updated_at

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "type": type,
                "entryType": entry_type,
                "input": input_,
            }
        )
        if id is not UNSET:
            field_dict["id"] = id
        if result is not UNSET:
            field_dict["result"] = result
        if status is not UNSET:
            field_dict["status"] = status
        if failure_reason is not UNSET:
            field_dict["failureReason"] = failure_reason
        if created_at is not UNSET:
            field_dict["created_at"] = created_at
        if updated_at is not UNSET:
            field_dict["updated_at"] = updated_at

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        type = DatasetType(d.pop("type"))

        entry_type = DatasetEntryType(d.pop("entryType"))

        input_ = cast(List[Any], d.pop("input"))

        id = d.pop("id", UNSET)

        result = cast(List[Any], d.pop("result", UNSET))

        _status = d.pop("status", UNSET)
        status: Union[Unset, DatasetStatus]
        if isinstance(_status, Unset):
            status = UNSET
        else:
            status = DatasetStatus(_status)

        failure_reason = d.pop("failureReason", UNSET)

        created_at = d.pop("created_at", UNSET)

        updated_at = d.pop("updated_at", UNSET)

        dataset = cls(
            type=type,
            entry_type=entry_type,
            input_=input_,
            id=id,
            result=result,
            status=status,
            failure_reason=failure_reason,
            created_at=created_at,
            updated_at=updated_at,
        )

        dataset.additional_properties = d
        return dataset

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import jaqpotpy.api.login as jaqlogin
import jaqpotpy.api.algorithms_api as alapi
import jaqpotpy.api.dataset_api as data_api
import jaqpotpy.api.models_api as models_api
import jaqpotpy.api.chempot_api as chempot_api
import jaqpotpy.api.task_api as task_api
import jaqpotpy.api.doa_api as doa_api
import jaqpotpy.helpers.jwt as jwtok
import jaqpotpy.helpers.helpers as help
import jaqpotpy.helpers.dataset_deserializer as ds
from jaqpotpy.api.types.api.model import create_model
from jaqpotpy.api.types.models import Model
from jaqpotpy.api.types.models.model_visibility import ModelVisibility
from jaqpotpy.api.types.models.feature import Feature
from jaqpotpy.api.types.client import AuthenticatedClient
from jaqpotpy.api.model_to_b64encoding import model_to_b64encoding
from jaqpotpy.helpers.logging import init_logger
import json
import jaqpotpy.api.feature_api as featapi
from jaqpotpy.helpers.serializer import JaqpotSerializer
from jaqpotpy.entities.dataset import Dataset
from jaqpotpy.entities.meta import MetaInfo
from jaqpotpy.entities.featureinfo import FeatureInfo
import pandas as pd
import numpy as np
import http.client as http_client
import getpass
import time
import jaqpotpy.doa.doa as jha
from sys import getsizeof
from tqdm import tqdm


ENCODING = 'utf-8'

# import matplotlib
# matplotlib.use('TkAgg')
# import matplotlib.pyplot as plt


class Jaqpot:
    """
    Deploys sklearn models on Jaqpot.

    Extended description of function.

    Parameters
    ----------
    base_url : The url on which Jaqpot services are deployed

    """

    def __init__(self, base_url=None, create_logs=False):
        # logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.INFO)
        self.log = init_logger(__name__, testing_mode=False, output_log_file=create_logs)
        if base_url:
            self.base_url = base_url
        else:
            self.base_url = 'https://app.appv2.jaqpot.org/'
        self.api_key = None
        self.user_id = None
        self.http_client = http_client

    def login(self, username, password):
        """
        Logins on Jaqpot.

        Parameters
        ----------
        username : username
        password : password

        """
        try:
            au_req = jaqlogin.authenticate_sync(self.base_url, username, password)
            self.api_key = au_req['authToken']
            # self.user_id = jwtok.decode_jwt(self.api_key).get('sub')
        except Exception as e:
            self.log.error("Could not login to jaqpot")
            # print("Error: " + str(e))

    def set_api_key(self, api_key):
        """
        Set's api key for authentication on Jaqpot.

        Parameters
        ----------
        api_key : api_key can be retireved from the application after logged in

        """
        self.api_key = api_key
        self.log.info("api key is set")

    def request_key(self, username, password):
        """
        Logins on Jaqpot.

        Parameters
        ----------
        username : username
        password : password

        """
        try:
            au_req = jaqlogin.authenticate_sync(self.base_url, username, password)
            self.api_key = au_req['authToken']
            self.log.info("api key is set")
            # self.user_id = jwtok.decode_jwt(self.api_key).get('sub')
        except Exception as e:
            self.log.error("Error: " + str(e))

    def request_key_safe(self):
        """
        Logins on Jaqpot by hiding the users password.

        """
        try:
            username = input("Username: ")
            password = getpass.getpass("Password: ")
            au_req = jaqlogin.authenticate_sync(self.base_url, username, password)
            self.api_key = au_req['authToken']
            self.log.info("api key is set")
            # self.user_id = jwtok.decode_jwt(self.api_key).get('sub')
        except Exception as e:
            self.log.error("Error: " + str(e))

    def my_info(self):
        """
        Prints user's info

        """
        self.log.info(jwtok.decode_jwt(self.api_key))
        return jwtok.decode_jwt(self.api_key)

    def get_algorithms(self, start=None, max=None):
        self.check_key()
        try:
            algos = alapi.get_allgorithms_sync(self.base_url, self.api_key, start, max)
            return algos
        except Exception as e:
            self.log.error("Error:" + str(e))

    def get_algorithms_classes(self, start=None, max=None):
        self.check_key()
        try:
            algos = alapi.get_allgorithms_classes(self.base_url, self.api_key, start, max)
            return algos
        except Exception as e:
            self.log.error("Error:" + str(e))

    def upload_dataset(self, df=None, id=None, title=None, description=None):
        if title is None:
            raise Exception("Please submit title of the dataset")
        if description is None:
            raise Exception("Please submit description of the dataset")
        df_titles = list(df)
        for t in df_titles:
            type_to_c = df[t].dtypes
            if type_to_c == 'int64':
                # print(type_to_c)
                df[t] = df[t].astype(float)
        # print(df.dtypes)
        df = df.replace(np.nan, '', regex=True)
        if type(df).__name__ != 'DataFrame':
            raise Exception("Cannot form a Jaqpot Dataset. Please provide a Dataframe")
        if id is not None:
            df.set_index(id, inplace=True)
            feat_titles = list(df)
            feats = []
            for f_t in feat_titles:
                fe = help.create_feature(f_t, self.user_id)
                jsonmi = json.dumps(fe, cls=JaqpotSerializer)
                # jsonmi = json.dumps(fe.__dict__)
                feats.append(jsonmi)
        else:
            feat_titles = list(df)
            feats = []
            for f_t in feat_titles:
                fe = help.create_feature(f_t, self.user_id)
                # jsonmi = json.dumps(fe.__dict__)
                jsonmi = json.dumps(fe, cls=JaqpotSerializer)
                feats.append(jsonmi)
        feat_map = {}
        featutes = []
        keyi = 0
        for feat in feats:
            feat_info = FeatureInfo()
            f = featapi.create_feature_sync(self.base_url, self.api_key, feat)
            feat_uri = self.base_url + "feature/" + f["_id"]
            feat_map[f["meta"]["titles"][0]] = keyi
            feat_info.uri = feat_uri
            feat_info.name = f["meta"]["titles"][0]
            feat_info.key = keyi
            featutes.append(feat_info.__dict__)
            keyi += 1
        dataset = Dataset()
        meta = MetaInfo()
        meta.creators = [self.user_id]
        if title is not None:
            meta.titles = [title]
        meta.descriptions = [description]
        dataset.meta = meta.__dict__
        dataset.totalRows = df.shape[0]
        dataset.totalColumns = df.shape[1]
        dataset.existence = "UPLOADED"
        data_entry = help.create_data_entry(df, feat_map, self.user_id)
        dataset.dataEntry = data_entry
        dataset.features = featutes
        jsondataset = json.dumps(dataset, cls=JaqpotSerializer)
        dataset_n = data_api.create_dataset_sync(self.base_url, self.api_key, jsondataset, self.log)
        self.log.info("Dataset created with id: " + dataset_n["_id"])
        return dataset_n["_id"]

    def chempot_predict(self, model, smiles, descriptors = "mordred", doa = False):
        """
        Makes a prediction from a Chempot model

        Parameters
        ----------
        model : string
            model is the id of the chempot model
        smiles : string
            The input SMILES.
        descriptors : string
            The descriptors of the model.
        doa : boolean
            Whether the Domain of Applicability will be taken into account
        
        Returns
        -------
        tuple
            The predicted dataset (first item) and the predicted feature (second item)
        """
        task = chempot_api.predict(self.base_url, self.api_key, model, smiles, descriptors, doa, self.log)
        
        percentange = 0
        taskid = task['_id']
        while percentange < 100:
            time.sleep(1)
            task = task_api.get_task(self.base_url, self.api_key, taskid)
            try:
                percentange = task['percentageCompleted']
            except KeyError:
                percentange = 0
            self.log.info("completed " + str(percentange))
        predictedDataset = task['resultUri']
        dar = predictedDataset.split("/")
        dataset = data_api.get_dataset(self.base_url, self.api_key, dar[len(dar)-1], self.log)
        df, predicts = ds.decode_predicted(dataset)
        return df, predicts


    def predict(self, df=None, modelId=None):
        df_titles = list(df)
        for t in df_titles:
            type_to_c = df[t].dtypes
            # if type_to_c == 'int64':
                # print(type_to_c)
                # df[t] = df[t].astype(float)
        # print(df.dtypes)
        df = df.replace(np.nan, '', regex=True)
        if type(df).__name__ != 'DataFrame':
            raise Exception("Cannot form a Jaqpot Dataset. Please provide a Dataframe")
        model = models_api.get_model(self.base_url, self.api_key, modelId, self.log)
        # feats = []
        # for featUri in model['independentFeatures']:
        #     featar = featUri.split("/")
        #     feat = featapi.get_feature(self.base_url, self.api_key, featar[len(featar)-1])
        #     feats.append(feat)
        # for featUri in model['additionalInfo']['independentFeatures']:
        #     print(featUri)
            # featar = featUri.split("/")
            # feat = featapi.get_feature(self.base_url, self.api_key, featar[len(featar)-1])
            # feats.append(feat)
        feat_map = {}
        featutes = []
        keyi = 0

        for key,value in model['additionalInfo']['independentFeatures'].items():
            feat_info = FeatureInfo()
            # f = featapi.create_feature_sync(self.base_url, self.api_key, feat)
            feat_uri = key
            feat_map[value] = keyi
            feat_info.uri = feat_uri
            feat_info.name = value
            feat_info.key = keyi
            featutes.append(feat_info.__dict__)
            keyi += 1

        dataset = Dataset()
        meta = MetaInfo()
        meta.creators = [self.user_id]
        dataset.meta = meta.__dict__
        dataset.totalRows = df.shape[0]
        dataset.totalColumns = df.shape[1]
        # dataset.existence = "UPLOADED"
        data_entry = help.create_data_entry(df, feat_map, self.user_id)
        dataset.dataEntry = data_entry
        dataset.features = featutes
        jsondataset = json.dumps(dataset, cls=JaqpotSerializer)
        dataset_n = data_api.create_dataset_sync(self.base_url, self.api_key, jsondataset, self.log)
        datasetId = dataset_n["_id"]
        datasetUri = self.base_url + "dataset/" + datasetId
        task = models_api.predict(self.base_url, self.api_key, dataseturi=datasetUri, modelid=modelId, logger=self.log)
        percentange = 0
        taskid = task['_id']
        while percentange < 100:
            time.sleep(1)
            task = task_api.get_task(self.base_url, self.api_key, taskid)
            try:
                percentange = task['percentageCompleted']
            except KeyError:
                percentange = 0
            self.log.info("completed " + str(percentange))
        predictedDataset = task['resultUri']
        dar = predictedDataset.split("/")
        dataset = data_api.get_dataset(self.base_url, self.api_key, dar[len(dar)-1], self.log)
        df, predicts = ds.decode_predicted(dataset)
        return df, predicts


    def deploy_SklearnModel(self, model, name, description, visibility):

        auth_client = AuthenticatedClient(base_url=self.base_url, token=self.api_key)
        actual_model = model_to_b64encoding(model.copy())
        body_model = Model(name = name, 
                            type=model.type, 
                            jaqpotpy_version=model.jaqpotpy_version,
                            libraries = model.libraries, 
                            dependent_features=[Feature(key=feature_i['key'], name=feature_i['name'], feature_type=feature_i['featureType']) for feature_i in model.dependentFeatures],
                            independent_features=[Feature(key=feature_i['key'], name=feature_i['name'], feature_type=feature_i['featureType']) for feature_i in model.independentFeatures],
                            visibility=ModelVisibility(visibility), 
                            actual_model=actual_model,
                            description = description)
        
        response = create_model.sync_detailed(client=auth_client, body=body_model)
        if response.status_code < 300:
            self.log.info("Model has been successfully uploaded. The url of the model is " + response.headers.get('Location'))
        else:
            error = response.headers.get('error')
            error_description = response.headers.get('error_description')
            self.log.error("Error code: " + str(response.status_code.value))


    def deploy_XGBoost(self, model, X, y, title, description, algorithm, doa=None):
        """
        Deploys XGBoost model to Jaqpot.

        Extended description of function.

        Parameters
        ----------
        model : XGBoost trained model
            model is a trained model that occurs from the XGBoost family of algorithms
        X : pandas dataframe
            The dataframe that is used to train the model (X variables).
        y : pandas dataframe
            The dataframe that is used to train the model (y variables).
        title: String
            The title of the model
        description: String
            The description of the model
        algorithm: String
            The algorithm that the model implements
        doa: pandas dataframe
            The dataset used to create the domain of applicability of the model

        Returns
        -------
        string
            The id of the model that uploaded

        """
        if isinstance(X, pd.DataFrame) is False:
            raise Exception('Function deploy_glm supports pandas dataframe or series. X is not one')
        if isinstance(y, pd.DataFrame) is False and isinstance(y, pd.Series) is False:
            raise Exception('Function deploy_glm supports pandas dataframe or series. Y is not one')
        # coef_flatten = model.coef_.flatten()
        # intercept_flatten = model.intercept_.flatten()
        # coef_df = pd.DataFrame(list(zip(list(X), coef_flatten)), columns=['Features', 'coeff'])
        # intercept = pd.DataFrame([('(intercept)', intercept_flatten[0])], columns=['Features', 'coeff'])
        # coefs_all = intercept.append(coef_df, ignore_index=True)
        # coefs = {}
        additionalInfo = {}
        # for key in coefs_all.values:
        #     coefs[key[0]] = key[1]
        # additionalInfo['coefficients'] = coefs
        additionalInfo['inputSeries'] = list(X)
        pretrained = help.create_pretrain_req(model, X, y, title, description,
                                              algorithm, "XGBoost model", "XGBoost-model",
                                              additionalInfo)
        # j = json.dumps(pretrained)
        j = json.dumps(pretrained, cls=JaqpotSerializer)
        if doa is None:
            response = models_api.post_pretrained_model(self.base_url, self.api_key, j, self.log)
            if response.status_code < 300:
                resp = response.json()
                self.log.info("Model with id: " + resp['modelId'] + " created. Please visit the application to proceed")
                return resp['modelId']
            else:
                resp = response.json()
                self.log.error("Some error occured: " + resp['message'])
                return
        else:
            response = models_api.post_pretrained_model(self.base_url, self.api_key, j, self.log)
            if response.status_code < 300:
                resp = response.json()
                self.log.info("Model with id: " + resp['modelId'] + " created. Storing Domain of applicability")
                modid = resp['modelId']
            else:
                resp = response.json()
                self.log.error("Some error occured: " + resp['message'])
                return
            # loop = asyncio.get_event_loop()
            # a = loop.create_task(jha.calculate_a(X))
            # b = loop.create_task(jha.calculate_doa_matrix(X))

            a = jha.calculate_a(X)
            b = jha.calculate_doa_matrix(X)
            # all_groups = asyncio.gather(a, b)
            # results = loop.run_until_complete(all_groups)
            doa = help.create_doa(inv_m=b.values.tolist(), a=a, modelid=resp['modelId'])
            j = json.dumps(doa, cls=JaqpotSerializer)
            resp = doa_api.post_models_doa(self.base_url, self.api_key, j, self.log)
            if resp == 201:
                self.log.info("Stored Domain of applicability. Visit the application to proceed")
                return modid

    def get_model_by_id(self, model):
        """
        Retrieves user's model by ID.

        Parameters
        ----------
        model : str
            The model's ID.
        
        
        Returns
        -------
        Object
            The particular model.

        """
        return models_api.get_model(self.base_url, self.api_key, model, self.log)

    def get_raw_model_by_id(self, model):
        """
        Retrieves raw model by ID.

        Parameters
        ----------
        model : str
            The model's ID.

        Returns
        Returns
        -------
        Object
            The particular model and the raw model.
        """

        # raw_model = models_api.get_raw_model(self.base_url, self.api_key, model, self.log)

        validating = jaqlogin.validate_api_key(self.base_url, self.api_key)
        if validating is not True:
            self.log.error(validating)
        else:
            return models_api.get_raw_model(self.base_url, self.api_key, model, self.log)

    def get_feature_by_id(self, feature):
        """
        Retrieves a Jaqpot feature.

        Parameters
        ----------
        feature : str
            The feature's ID.
        
        
        Returns
        -------
        Object
            The particular feature.

        """
        return featapi.get_feature(self.base_url, self.api_key, feature)

    def get_my_models(self, minimum, maximum):
        """
        Retrieves user's models.

        Parameters
        ----------
        minimum : int
            The index of the first model.
        maximum : int
            The index of the last model.
        
        Returns
        -------
        Object
            The models of the user.

        """
        return models_api.get_my_models(self.base_url, self.api_key, minimum, maximum, self.log)

    
    def get_orgs_models(self, organization, minimum, maximum):
        """
        Retrieves organization's models.

        Parameters
        ----------
        organization: str
            The organization's ID.
        minimum : int
            The index of the first model.
        maximum : int
            The index of the last model.
        
        Returns
        -------
        Object
            The models of the organization.

        """
        return models_api.get_orgs_models(self.base_url, self.api_key, organization, minimum, maximum, self.log)


    def get_models_by_tag(self, tag, minimum, maximum):
        """
        Retrieves models with a particular tag.

        Parameters
        ----------
        tag: str
            The model's tag.
        minimum : int
            The index of the first model.
        maximum : int
            The index of the last model.
        
        Returns
        -------
        Object
            The models of the organization.

        """
        return models_api.get_models_by_tag(self.base_url, self.api_key, tag, minimum, maximum, self.log)

    def get_models_by_tag_and_org(self, organization, tag, minimum, maximum):
        """
        Retrieves models of an organization with a particular tag.

        Parameters
        ----------
        organization: str:
            The organization's ID.
        tag: str
            The model's tag.
        minimum : int
            The index of the first model.
        maximum : int
            The index of the last model.
        
        Returns
        -------
        Object
            The models of the organization.

        """
        return models_api.get_models_by_tag(self.base_url, self.api_key, organization, tag, minimum, maximum, self.log)


    def get_dataset(self, dataset):
        """
        Retrieves a dataset.

        Parameters
        ----------
        dataset: str
            The dataset's ID.
        
        Returns
        -------
        pd.DataFrame()
            DataFrame of the dataset.

        Note
        ----
        Mazimum rows retrieved: 1000.
        """

        dataRetrieved = data_api.get_dataset(self.base_url, self.api_key, dataset, self.log)

        feats = ["" for i in range(len(dataRetrieved['features']))]
        
        for item in dataRetrieved['features']:
            feats[int(item['key'])] = item['name']

        indices = [item['entryId']['name'] for item in dataRetrieved['dataEntry']]

        data = [[item['values'][str(feats.index(col))] for col in feats] for item in dataRetrieved['dataEntry']]

        return pd.DataFrame(data,index=indices, columns=feats) 

    def get_doa(self,model):
        """
        Retrieves model's domain of applicability.

        Parameters
        ----------
        model: str
            The model's ID.
        
        Returns
        -------
        Object
            The model's domain of applicability.

        """
        return doa_api.get_models_doa(self.base_url, self.api_key, model, self.log)



from rdkit import Chem
import torch
from torch_geometric.data import Data
import warnings
import pickle
import json
from abc import ABC, abstractmethod

class Featurizer(ABC):
    """
    Abstract base class for featurizers.
    """

    def __call__(self, *args, **kwargs):
        """
        Featurizes the input data.
        
        Returns:
            The featurized data.
        """
        return self.featurize(*args, **kwargs)
    
    @abstractmethod
    def featurize(self, *args, **kwargs):
        """
        Abstract method to featurize the input data.

        Returns:
            The featurized data.
        """
        pass

class SmilesGraphFeaturizer(Featurizer):
    """
    Featurizes SMILES strings into graph data suitable for graph neural networks.

    Attributes:
        include_edge_features (bool): Whether to include edge features in the featurization.
        warnings_enabled (bool): Whether to enable warnings.
        atom_allowable_sets (dict): Allowable sets for atom features.
        bond_allowable_sets (dict): Allowable sets for bond features.
    """
    # Dictionary mapping atom features to their respective featurization function
    SUPPORTED_ATOM_FEATURES = {"symbol": lambda atom: atom.GetSymbol(),
                               "degree": lambda atom: atom.GetDegree(),
                               "total_degree": lambda atom: atom.GetTotalDegree(),
                               "formal_charge": lambda atom: atom.GetFormalCharge(),
                               "num_radical_electrons": lambda atom: atom.GetNumRadicalElectrons(),
                               "hybridization": lambda atom: atom.GetHybridization(),
                               "is_aromatic": lambda atom: atom.GetIsAromatic(),
                               "is_in_ring": lambda atom: atom.IsInRing(),
                               "total_num_hs": lambda atom: atom.GetTotalNumHs(),
                               "num_explicit_hs": lambda atom: atom.GetNumExplicitHs(),
                               "num_implicit_hs": lambda atom: atom.GetNumImplicitHs(),
                               "_ChiralityPossible": lambda atom: atom.HasProp('_ChiralityPossible'),
                               "isotope": lambda atom: atom.GetIsotope(),
                               "total_valence": lambda atom: atom.GetTotalValence(),
                               "explicit_valence": lambda atom: atom.GetExplicitValence(),
                               "implicit_valence": lambda atom: atom.GetImplicitValence(),
                               "chiral_tag": lambda atom: atom.GetChiralTag(), 
                               "mass": lambda atom: (atom.GetMass()-14.5275)/9.4154}
    # Dictionary mapping bond features to their respective featurization function
    SUPPORTED_BOND_FEATURES = {"bond_type": lambda bond: bond.GetBondType(),
                               "is_conjugated": lambda bond: bond.GetIsConjugated(),
                               "is_in_ring": lambda bond: bond.IsInRing(),
                               "stereo": lambda bond: bond.GetStereo()}
    # List of features that should not be one-hot encoded
    NON_ONE_HOT_ENCODED = [
                           "formal_charge",
                           "num_radical_electrons",
                           "is_aromatic",
                           "_ChiralityPossible",
                           "is_conjugated",
                           "is_in_ring",
                           "mass"]
    
    def __init__(self, include_edge_features=True, warnings_enabled=True):
        """
        The SmilesGraphFeaturizer constructor.

        Args:
            include_edge_features (bool): Whether to include edge features in the featurization.
            warnings_enabled (bool): Whether to enable warnings.
        ```
        """
        
        self.include_edge_features = include_edge_features
        
        self.warnings_enabled = warnings_enabled
        
        self.atom_allowable_sets = {}
        self.bond_allowable_sets = {}
    
    def get_supported_atom_features(self):
        """Returns the names of the supported atom features."""
        return self.SUPPORTED_ATOM_FEATURES.keys()
    
    def get_supported_bond_features(self):
        """Returns the names of the supported bond features."""
        return self.SUPPORTED_BOND_FEATURES.keys()
    
    def config_from_other_featurizer(self, featurizer):
        """
        Configures the featurizer from another SmilesGraphFeaturizer instance.

        Args:
            featurizer (Featurizer): Another SmilesGraphFeaturizer instance.

        Returns:
            SmilesGraphFeaturizer: The configured featurizer.
        """
        
        self.include_edge_features = featurizer.include_edge_features
        
        self.warnings_enabled = featurizer.warnings_enabled
        
        self.set_atom_allowable_sets(featurizer.atom_allowable_sets)

        self.set_bond_allowable_sets(featurizer.bond_allowable_sets)
        
        return self
        
    def set_atom_allowable_sets(self, atom_allowable_sets_dict):
        """
        Sets the allowable sets for atom feature.

        Args:
            atom_allowable_sets_dict (dict): Dictionary of allowable sets per atom feature.
        """

        self.atom_allowable_sets = dict()
        for atom_feature, allowable_set in atom_allowable_sets_dict.items():
            self.add_atom_feature(atom_feature, allowable_set)
    
    def set_bond_allowable_sets(self, bond_allowable_sets_dict):
        """
        Sets the allowable sets for bond features.

        Args:
            bond_allowable_sets_dict (dict): Dictionary of allowable sets per bond feature.
        """
        
        self.bond_allowable_sets = dict()
        for bond_feature, allowable_set in bond_allowable_sets_dict.items():
            self.add_bond_feature(bond_feature, allowable_set)

    def add_atom_feature(self, atom_feature, allowable_set=None):
        """
        Adds an atom feature to the featurizer.

        Args:
            atom_feature (str): The atom feature name to be added.
            allowable_set (list or None): The allowable set for the atom feature.
        
        Example:
        ```
        >>> featurizer = SmilesGraphFeaturizer()
        >>> featurizer.add_atom_feature('symbol', ['C', 'O', 'N', 'Cl']) # categorical characteristics need allowable set of categories
        >>> featurizer.add_atom_feature('is_in_ring') # allowable set is set to Nonefor non-categorical
        >>> featurizer.atom_allowable_sets
        {'symbol': ['C', 'O', 'N', 'Cl'], 'is_in_ring': None}
        ```
        Notes:
            'UNK' is used as a special keyword. When an atom feature is encoded and an unknown 
            value (not in the allowable set) is encountered during featurization, it will be encoded as 'UNK'.
            This allows handling of unexpected values by mapping them to a default or placeholder category.
        """
        # The name of the atom feature must be exactly as given by RDKit
        if atom_feature not in self.get_supported_atom_features():
            raise ValueError(f"Unsupported atom feature '{atom_feature}'")
    
        if atom_feature in self.atom_allowable_sets.keys():
            self.warning(f"The atom allowable set for '{atom_feature}' will be overwritten.")

        if atom_feature in self.NON_ONE_HOT_ENCODED:
            if allowable_set is not None:
                self.warning(f"Atom allowable set given for '{atom_feature}' will be ignored (not one-hot encoded)")
            # Placeholder for the feature to contain a single element only
            self.atom_allowable_sets[atom_feature] = None 
        else:
            if allowable_set is None:
                self.warning(f"The atom allowable set for '{atom_feature}' is set to default.")
                # This is the for the one hot encoded features
                allowable_set = self.get_default_atom_allowable_set(atom_feature)
            elif not isinstance(allowable_set, (list, tuple)):
                raise TypeError("Input dictionary must have values of type list, tuple or None.")
            # Creates the feature list that will be transfored in one-hot
            self.atom_allowable_sets[atom_feature] = list(allowable_set)

    def add_bond_feature(self, bond_feature, allowable_set=None):
        """
        Adds a bond feature to the featurizer.

        Args:
            bond_feature (str): The bond feature to add.
            allowable_set (list or None): The allowable set for the bond feature.

        Example:
        ```
        >>> from rdkit import Chem
        >>> featurizer = SmilesGraphFeaturizer()
        >>> featurizer.add_bond_feature('bond_type', [Chem.rdchem.BondType.SINGLE, Chem.rdchem.BondType.DOUBLE, Chem.rdchem.BondType.TRIPLE])
        >>> featurizer.add_bond_feature('is_conjugated')
        >>> featurizer.bond_allowable_sets
        {'bond_type': [rdkit.Chem.rdchem.BondType.SINGLE,
        rdkit.Chem.rdchem.BondType.DOUBLE,
        rdkit.Chem.rdchem.BondType.TRIPLE],
        'is_conjugated': None}
        ```

        Notes:
            'UNK' is used as a special keyword. When a bond feature is encoded and an unknown 
            value (not in the allowable set) is encountered during featurization, it will be encoded as 'UNK'.
            This allows handling of unexpected values by mapping them to a default or placeholder category.
        """

        if bond_feature not in self.get_supported_bond_features():
            raise ValueError(f"Unsupported bond feature '{bond_feature}'")
        
        if bond_feature in self.bond_allowable_sets.keys():
            self.warning(f"The bond allowable set for '{bond_feature}' will be overwritten.")

        if bond_feature in self.NON_ONE_HOT_ENCODED:
            if allowable_set is not None:
                self.warning(f"Bond allowable set given for '{bond_feature}' will be ignored (not one-hot encoded)")
            self.bond_allowable_sets[bond_feature] = None
        else:
            if allowable_set is None:
                self.warning(f"The bond allowable set for '{bond_feature}' is set to default.")
                # This is the for the one hot encoded features
                allowable_set = self.get_default_bond_allowable_set(bond_feature)
            if not isinstance(allowable_set, (list, tuple)):
                raise TypeError("Input dictionary must have values of type list, tuple or None.")
            # Creates the feature list that will be transfored in one-hot
            self.bond_allowable_sets[bond_feature] = list(allowable_set)
    
    def get_atom_feature_labels(self):
        """
        Returns the labels for the atom features.

        Returns:
            list: The labels for the atom features.
        """
        
        atom_feature_labels = []
        
        for feature in self.atom_allowable_sets.keys():
            if feature in self.NON_ONE_HOT_ENCODED:
                atom_feature_labels.append(feature)
            else:
                atom_feature_labels += [f"{feature}.{value}" for value in self.atom_allowable_sets[feature]]
        
        return atom_feature_labels
    
    
    def get_bond_feature_labels(self):
        """
        Returns the labels for the bond features.

        Returns:
            list: The labels for the bond features.
        """
        
        bond_feature_labels = []
        
        for feature in self.bond_allowable_sets.keys():
            if feature in self.NON_ONE_HOT_ENCODED:
                bond_feature_labels.append(feature)
            else:
                bond_feature_labels += [f"{feature}.{value}" for value in self.bond_allowable_sets[feature]]
        
        return bond_feature_labels

    def get_default_atom_allowable_set(self, atom_feature):
        """
        Returns the default allowable set for an atom feature.

        Args:
            atom_feature (str): The atom feature.

        Returns:
            list: The default allowable set for the atom characteristic.
       
        """
        match atom_feature:
            case "symbol":
                return ['C', 'O', 'N', 'Cl', 'S', 'F', 'Na', 'P', 'Br', 'Si', 'K', 'Sn', 'UNK'],
            case "degree":
                return [2, 1, 3, 4, 0]
            case "hybridization":
                return [Chem.rdchem.HybridizationType.SP2,
                        Chem.rdchem.HybridizationType.SP3,
                        Chem.rdchem.HybridizationType.S,
                        Chem.rdchem.HybridizationType.SP]
            case "total_num_hs":
                return [0, 1, 2, 3]
            case "implicit_valence":
                return [0, 1, 2, 3, 4]
            case "explicit_valence":
                return [4, 2, 3, 1, 0, 'UNK']
            #case "formal_charge":
            #    return [0, -1, 1, 'UNK']
            case _:
                raise ValueError(f"No default allowable set for atom characteristic '{atom_feature}'. You must set your own allowable set.")
    
    def get_default_bond_allowable_set(self, bond_feature):
        """
        Returns the default allowable set for a bond feature.

        Args:
            bond_feature (str): The bond feature.

        Returns:
            list: The default allowable set for the bond feature.
        """

        match bond_feature:
            case "bond_type":
                return [Chem.rdchem.BondType.SINGLE,
                        Chem.rdchem.BondType.DOUBLE,
                        Chem.rdchem.BondType.TRIPLE,
                        Chem.rdchem.BondType.AROMATIC]
            case _:
                raise ValueError(f"No default allowable set for bond characteristic '{bond_feature}'. You must set your own allowable set.")

    def set_default_config(self):
        """
        Sets the default configuration for the featurizer.
        """

        atom_allowable_sets = {
            "symbol": self.get_default_atom_allowable_set("symbol"),
            "degree": self.get_default_atom_allowable_set("degree"),
            "hybridization": self.get_default_atom_allowable_set("hybridization"),
            "total_num_hs": self.get_default_atom_allowable_set("total_num_hs"),
            "implicit_valence": self.get_default_atom_allowable_set("implicit_valence"),
            "explicit_valence": self.get_default_atom_allowable_set("explicit_valence"),
            "is_aromatic": None,
            "_ChiralityPossible": None,
            "formal_charge": None,
        }

        bond_allowable_sets = {
            "bond_type": self.get_default_bond_allowable_set("bond_type"),
            "is_conjugated": None,
            "is_in_ring": None,
        }

        self.set_atom_allowable_sets(atom_allowable_sets)

        if self.include_edge_features:
            self.set_bond_allowable_sets(bond_allowable_sets)
    
    def extract_molecular_features(self, mol):
        """
        Extracts molecular features from a molecule.

        Args:
            mol (rdkit.Chem.Mol): The molecule to extract features from.

        Returns:
            tuple: The atom features and bond features as tensors.
        """
        
        mol_atom_features = []
        for atom in mol.GetAtoms():
            mol_atom_features.append(self.atom_features(atom))
        mol_atom_features = torch.tensor(mol_atom_features, dtype=torch.float32)
        
        if not self.include_edge_features:
            return mol_atom_features, None
        
        mol_bond_features = []
        for bond in mol.GetBonds():
            mol_bond_features.append(self.bond_features(bond))
            mol_bond_features.append(self.bond_features(bond)) # do twice (undirectional graph)
        mol_bond_features = torch.tensor(mol_bond_features, dtype=torch.float32)
        
        return mol_atom_features, mol_bond_features
        
        
    def atom_features(self, atom):
        """
        Extracts features for an atom.

        Args:
            atom (rdkit.Chem.Atom): The atom to extract features from.

        Returns:
            list: The features for the atom.
        """
        
        feats = []
        for feature in self.atom_allowable_sets.keys():
            # Contains the rdkit function for feature calculation
            property_getter = self.SUPPORTED_ATOM_FEATURES[feature]
            feat = property_getter(atom)
            
            if feature in self.NON_ONE_HOT_ENCODED:
                feats.append(feat)
            else:
                allowable_set = self.atom_allowable_sets[feature]
                if 'UNK' in allowable_set:
                    one_hot_encoded_feat = self.one_of_k_encoding_unk(feat, allowable_set)
                else:
                    one_hot_encoded_feat = self.one_of_k_encoding(feat, allowable_set, feature_name=feature)
                feats.extend(one_hot_encoded_feat)
        return feats
    
    def bond_features(self, bond):
        """
        Extracts features for a bond.

        Args:
            bond (rdkit.Chem.Bond): The bond to extract features from.

        Returns:
            list: The features for the bond.
        """

        feats = []
        for feature in self.bond_allowable_sets.keys():
            property_getter = self.SUPPORTED_BOND_FEATURES[feature]
            feat = property_getter(bond)
            
            if feature in self.NON_ONE_HOT_ENCODED:
                feats.append(feat)
            else:
                allowable_set = self.bond_allowable_sets[feature]
                if 'UNK' in allowable_set:
                    one_hot_encoded_feat = self.one_of_k_encoding_unk(feat, allowable_set)
                else:
                    one_hot_encoded_feat = self.one_of_k_encoding(feat, allowable_set, feature_name=feature)
                feats.extend(one_hot_encoded_feat)
        return feats

    def one_of_k_encoding(self, x, allowable_set, feature_name=None):
        """
        One-hot encodes a value based on an allowable set.

        Args:
            x: The value to encode.
            allowable_set (list): The allowable set for the value.
            characteristic_name (str): The name of the characteristic (optional).

        Returns:
            list: The one-hot encoded vector.
        """
        if x not in allowable_set:
            feature_text = f"{feature_name} " if feature_name else ""
            self.warning(f"Ignoring input {feature_text}{x}, not in allowable set {allowable_set}")
        return [x == s for s in allowable_set]

    def one_of_k_encoding_unk(self, x, allowable_set):
        """
        One-hot encodes a value based on an allowable set, mapping unknown values to 'UNK'.

        Args:
            x: The value to encode.
            allowable_set (list): The allowable set for the value.

        Returns:
            list: The one-hot encoded value.
        """
        if x not in allowable_set:
            x = 'UNK'
        return [x == s for s in allowable_set]
    
    def adjacency_matrix(self, mol):
        """
        Computes the adjacency matrix for a molecule.

        Args:
            mol (rdkit.Chem.Mol): The molecule to compute the adjacency matrix for.

        Returns:
            torch.Tensor: The adjacency matrix.
        """
        ix1, ix2 = [], []
        for bond in mol.GetBonds():
            start, end = bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()
            ix1 += [start, end]
            ix2 += [end, start]
        adj_norm = torch.asarray([ix1, ix2], dtype=torch.int64) # Needs to be in COO Format
        return adj_norm
        
    def get_num_node_features(self):
        """
        Returns the number of node features.

        Returns:
            int: The number of node features.
        """
        return len(self.get_atom_feature_labels())
    

    def get_num_edge_features(self):
        """
        Returns the number of edge features.

        Returns:
            int: The number of edge features.
        """
        return len(self.get_bond_feature_labels())
    
    
    def warning(self, message):
        """
        Issues a warning if warnings are enabled.

        Args:
            message (str): The warning message.
        """
        if self.warnings_enabled:
            warnings.warn(message)
    
    def enable_warnings(self, enable=True):
        """
        Enables or disables warnings.

        Args:
            enable (bool): Whether to enable warnings.
        """
        self.warnings_enabled = enable
    
    def featurize(self, sm, y=None):
        """
        Featurizes a SMILES string into graph data.

        Args:
            sm (str): The SMILES string.
            y: The target value (optional).

        Returns:
            torch_geometric.data.Data: The featurized data.
        """
        
        mol = Chem.MolFromSmiles(sm)
        adjacency_matrix = self.adjacency_matrix(mol)
        mol_atom_features, mol_bond_features = self.extract_molecular_features(mol)
        
        return Data(x=mol_atom_features,
                    edge_index=adjacency_matrix,
                    edge_attr=mol_bond_features,
                    y=y,
                    smiles=sm)
        
    
    def save_config(self, config_file="featurizer_config.pkl"):
        """
        Saves the featurizer configuration to a file.

        Args:
            config_file (str): The file to save the configuration to.
        """
        config = {
            'warnings_enabled': self.warnings_enabled,
            'include_edge_features': self.include_edge_features,
            'atom_allowable_sets': self.atom_allowable_sets,
            'bond_allowable_sets': self.bond_allowable_sets,
        }

        with open(config_file, 'wb') as f:
            pickle.dump(config, f)

        
    def load_config(self, config_file):
        """
        Loads the featurizer configuration from a file.

        Args:
            config_file (str): The file to load the configuration from.

        Returns:
            SmilesGraphFeaturizer: The configured featurizer.
        """

        with open(config_file, 'rb') as f:
            config = pickle.load(f)

        self.warnings_enabled = config['warnings_enabled']
        self.include_edge_features = config['include_edge_features']

        self.set_atom_allowable_sets(config['atom_allowable_sets'])

        self.set_bond_allowable_sets(config['bond_allowable_sets'])

        return self
    

    def save(self, filepath='featurizer.pkl'):
        """
        Saves the featurizer to a file.

        Args:
            filepath (str): The file to save the featurizer to.
        """
        with open(filepath, "wb") as file:
            pickle.dump(self, file)

    def __repr__(self):
        """
        Returns the string representation of the featurizer.

        Returns:
            str: The string representation of the featurizer.
        """
        attributes = {
            'atom_allowable_sets': self.atom_allowable_sets,
            'bond_allowable_sets': self.bond_allowable_sets,
        }
        return type(self).__name__ + '(' + json.dumps(attributes, indent=4) + ')'
    
    def __str__(self):
        """
        Returns the string representation of the featurizer.

        Returns:
            str: The string representation of the featurizer.
        """
        return self.__repr__()



from enum import Enum


class DatasetStatus(str, Enum):
    CREATED = "CREATED"
    EXECUTING = "EXECUTING"
    FAILURE = "FAILURE"
    SUCCESS = "SUCCESS"

    def __str__(self) -> str:
        return str(self.value)



from enum import Enum


class ModelType(str, Enum):
    R_BNLEARN_DISCRETE = "R_BNLEARN_DISCRETE"
    R_CARET = "R_CARET"
    R_GBM = "R_GBM"
    R_NAIVE_BAYES = "R_NAIVE_BAYES"
    R_PBPK = "R_PBPK"
    R_RF = "R_RF"
    R_RPART = "R_RPART"
    R_SVM = "R_SVM"
    R_TREE_CLASS = "R_TREE_CLASS"
    R_TREE_REGR = "R_TREE_REGR"
    SKLEARN = "SKLEARN"
    TORCH = "TORCH"

    def __str__(self) -> str:
        return str(self.value)



"""
Tests for Docking
"""
import os
import platform
import unittest
#import pytest
import logging
import numpy as np
import jaqpotpy as jp
from jaqpotpy.descriptors.base_classes import ComplexFeaturizer
from jaqpotpy.models import Model
from jaqpotpy.docking.pose_generation import PoseGenerator

IS_WINDOWS = platform.system() == 'Windows'


class TestDockingUtils(unittest.TestCase):
  """
  Does sanity checks on pose generation.
  """

  def setUp(self):
    current_dir = os.path.dirname(os.path.realpath(__file__))
    # self.protein_file = os.path.join(current_dir, "1jld_protein.pdb")
    # self.protein_file = os.path.join(current_dir, "7zb6.pdb")
    # self.ligand_file = os.path.join(current_dir, "mulno.sdf")
    # self.ligand_file = os.path.join(current_dir, "1jld_ligand.sdf")

    self.protein_file = os.path.join(current_dir, "1a9m_pocket.pdb")
    self.ligand_file = os.path.join(current_dir, "1a9m_ligand.sdf")


  #@pytest.mark.slow
  @unittest.skip("Docking has not been integrated in the latest jaqpotpy version")
  def test_docker_init(self):
    """Test that Docker can be initialized."""
    vpg = jp.docking.VinaPoseGenerator()
    jp.docking.Docker(vpg)

  @unittest.skip("test only for conda env")
  def test_docking_utils(self):
      from jaqpotpy.docking.utils import create_hydrated_pdbqt_pdb
      protein_pdbqt = create_hydrated_pdbqt_pdb(self.protein_file, "./tmp")
      print(protein_pdbqt)


  @unittest.skipIf(IS_WINDOWS, "vina is not supported in windows")
  @unittest.skip("Docking has not been integrated in the latest jaqpotpy version")
  #@pytest.mark.slow
  def test_docker_dock(self):
    """Test that Docker can dock."""
    # We provide no scoring model so the docker won't score
    vpg = jp.docking.VinaPoseGenerator()
    docker = jp.docking.Docker(vpg)
    docked_outputs = docker.dock((self.protein_file, self.ligand_file),
                                 exhaustiveness=10,
                                 num_modes=10,
                                 out_dir="./tmp")

    print(docked_outputs)
    # Check only one output since num_modes==1
    assert len(list(docked_outputs)) == 10



"""Contains all the data models used in inputs/outputs"""

from .create_invitations_body import CreateInvitationsBody
from .dataset import Dataset
from .dataset_csv import DatasetCSV
from .dataset_csv_status import DatasetCSVStatus
from .dataset_entry_type import DatasetEntryType
from .dataset_status import DatasetStatus
from .dataset_type import DatasetType
from .error_code import ErrorCode
from .error_response import ErrorResponse
from .feature import Feature
from .feature_feature_dependency import FeatureFeatureDependency
from .feature_meta import FeatureMeta
from .feature_meta_additional_property import FeatureMetaAdditionalProperty
from .feature_type import FeatureType
from .get_models_response_200 import GetModelsResponse200
from .get_shared_models_response_200 import GetSharedModelsResponse200
from .library import Library
from .model import Model
from .model_meta import ModelMeta
from .model_meta_additional_property import ModelMetaAdditionalProperty
from .model_type import ModelType
from .model_visibility import ModelVisibility
from .organization import Organization
from .organization_invitation import OrganizationInvitation
from .organization_invitation_status import OrganizationInvitationStatus
from .organization_visibility import OrganizationVisibility
from .partially_update_model_body import PartiallyUpdateModelBody
from .partially_update_model_feature_body import PartiallyUpdateModelFeatureBody
from .search_models_response_200 import SearchModelsResponse200
from .user import User

__all__ = (
    "CreateInvitationsBody",
    "Dataset",
    "DatasetCSV",
    "DatasetCSVStatus",
    "DatasetEntryType",
    "DatasetStatus",
    "DatasetType",
    "ErrorCode",
    "ErrorResponse",
    "Feature",
    "FeatureFeatureDependency",
    "FeatureMeta",
    "FeatureMetaAdditionalProperty",
    "FeatureType",
    "GetModelsResponse200",
    "GetSharedModelsResponse200",
    "Library",
    "Model",
    "ModelMeta",
    "ModelMetaAdditionalProperty",
    "ModelType",
    "ModelVisibility",
    "Organization",
    "OrganizationInvitation",
    "OrganizationInvitationStatus",
    "OrganizationVisibility",
    "PartiallyUpdateModelBody",
    "PartiallyUpdateModelFeatureBody",
    "SearchModelsResponse200",
    "User",
)



"""
Tests for Docking
"""
import os
import platform
import unittest
import logging
#import pytest
import numpy as np
import jaqpotpy as jp
from jaqpotpy.descriptors.base_classes import ComplexFeaturizer
from jaqpotpy.models import Model
from jaqpotpy.docking.pose_generation import PoseGenerator
import rdkit.Chem
import rdkit.Chem
#from jaqpotpy.docking.utils import create_hydrated_pdbqt_pdb
# pylint: disable=no-member

IS_WINDOWS = platform.system() == 'Windows'


class TestDocking(unittest.TestCase):
  """
  Does sanity checks on pose generation.
  """

  def setUp(self):
    current_dir = os.path.dirname(os.path.realpath(__file__))
    # self.protein_file = os.path.join(current_dir, "1jld_protein.pdb")
    # self.protein_file = os.path.join(current_dir, "7zb6.pdb")
    # self.protein_file = os.path.join(current_dir, "0_a.pdb")
    self.protein_file = os.path.join(current_dir, "117_a.pdb")
    self.ligand_file = os.path.join(current_dir, "mulno.sdf")
    # self.ligand_file = os.path.join(current_dir, "1jld_ligand.sdf")

    # self.protein_file = os.path.join(current_dir, "1a9m_pocket.pdb")
    # self.ligand_file = os.path.join(current_dir, "1a9m_ligand.sdf")
    # self.ligand_file = os.path.join(current_dir, "ZINC000787318646.sdf")

  #@pytest.mark.slow
  @unittest.skip("Docking has not been integrated in the latest jaqpotpy version")
  def test_docker_init(self):
    """Test that Docker can be initialized."""
    vpg = jp.docking.VinaPoseGenerator()
    jp.docking.Docker(vpg)

  @unittest.skip("Docking has not been integrated in the latest jaqpotpy version")
  def test_tdc(self):
    from jaqpotpy.datasets import TorchGraphDataset
    from jaqpotpy.models import Evaluator
    import torch
    from jaqpotpy.models import GCN_V1, MolecularTorchGeometric, AttentiveFP

    from jaqpotpy.descriptors.molecular import AttentiveFPFeaturizer

    from sklearn.metrics import accuracy_score, f1_score, roc_auc_score

    #from tdc.single_pred import ADME
    #data = ADME(name='BBB_Martins')
    data = None

    split = data.get_split()
    featurizer = AttentiveFPFeaturizer()
    train_dataset = TorchGraphDataset(smiles=split['train']['Drug'], y=split['train']['Y'], task='classification'
                                , featurizer=featurizer)

    test_dataset = TorchGraphDataset(smiles=split['test']['Drug'], y=split['test']['Y'], task='classification'
                                , featurizer=featurizer)

    train_dataset.create()
    test_dataset.create()
    val = Evaluator()
    val.dataset = test_dataset
    val.register_scoring_function('Accuracy', accuracy_score)
    val.register_scoring_function('F1', f1_score)
    val.register_scoring_function('Roc Auc', roc_auc_score)

    model = AttentiveFP(in_channels=39, hidden_channels=100, out_channels=2, edge_dim=10, num_layers=6,
                        num_timesteps=2).jittable()

    optimizer = torch.optim.Adam(model.parameters(), lr=0.0001, weight_decay=5e-4)
    criterion = torch.nn.CrossEntropyLoss()
    m = MolecularTorchGeometric(dataset=train_dataset
                                , model_nn=model, eval=val
                                , train_batch=262, test_batch=200
                                , epochs=840, optimizer=optimizer, criterion=criterion, device="cpu").fit()

    m.eval()

    mol_m = m.create_molecular_model()
    mol_m.model_title = "BBB_MODEL"
    mol_m.save()
  
  @unittest.skip("Docking has not been integrated in the latest jaqpotpy version")
  def test_smi_file(self):
      from rdkit import Chem
      from rdkit.Chem import AllChem
      suppl = Chem.SmilesMolSupplier("/Users/pantelispanka/Downloads/bbb_v01.smi")
      for mol in suppl:
          id = mol.GetProp('id')
          smile = Chem.MolToSmiles(mol)
          mol = Chem.AddHs(mol)
          AllChem.EmbedMolecule(mol, randomSeed=0xf00d)
          Chem.MolToMolBlock(mol)
          # ligand = "./" + str(id) + "_a_syn_ligand.sdf"
          ligand = "/Users/pantelispanka/Jaqpot/jaqpotpy/jaqpotpy/docking/test/bbb_v01_0_a_syn_ligand.sdf"
          # with Chem.SDWriter(ligand) as w:
          #     w.write(mol)
          vpg = jp.docking.VinaPoseGenerator(prepare_protein=True)
          docker = jp.docking.Docker(vpg)
          docked_outputs = docker.dock((self.protein_file, ligand),
                                       centroid=(80, 38, 48),
                                       box_dims=(12, 12, 12),
                                       exhaustiveness=12,
                                       num_modes=1,
                                       out_dir="./tmp",
                                       use_pose_generator_scores=True)

          # Check only one output since num_modes==1
          docked_outputs = list(docked_outputs)
          print(docked_outputs)
          print(len(docked_outputs))
          print(len(docked_outputs[0]))
          assert len(docked_outputs) == 1
          assert len(docked_outputs[0]) == 2

  @unittest.skip("Docking has not been integrated in the latest jaqpotpy version")
  def test_jaqpotpy(self):
    import warnings
    warnings.filterwarnings("ignore")
    import pandas as pd
    from rdkit import Chem
    from jaqpotpy import Jaqpot
    from jaqpotpy.models import MolecularModel
    jaqpot = Jaqpot()
    jaqpot.request_key("pantelispanka@gmail.com", "kapan2")
    model = MolecularModel().load_from_jaqpot(jaqpot, "oZZfU6RQgLnmHgk88hnc")
    model.save()
    smiles_file = "/Users/pantelispanka/guacamol_v1_all.csv"
    df = pd.read_csv(smiles_file)
    for index, smile in enumerate(df[:100000]['SMILES']):
        print(index)
        print(smile)
        mol = Chem.MolFromSmiles(smile)
        try:
            model(mol)
            print(model.prediction)
            print(model.probability)
        except Exception as e:
            print(str(e))

  # @unittest.skip("skipping automated test")
  @unittest.skipIf(IS_WINDOWS, "vina is not supported in windows")
  @unittest.skip("Docking has not been integrated in the latest jaqpotpy version")
  #@pytest.mark.slow
  def test_docker_dock(self):
    """Test that Docker can dock."""
    # We provide no scoring model so the docker won't score
    vpg = jp.docking.VinaPoseGenerator(calc_charges=False, add_hydrogens=False)
    docker = jp.docking.Docker(vpg)
    docked_outputs = docker.dock((self.protein_file, self.ligand_file),
                                 exhaustiveness=12,
                                 num_modes=4,
                                 out_dir="./tmp",
                                 use_pose_generator_scores=True)

    print(docked_outputs)
    print(list(docked_outputs))
    # Check only one output since num_modes==1
    assert len(list(docked_outputs)) == 4

  # @unittest.skip("skipping automated test")
  @unittest.skipIf(IS_WINDOWS, "vina is not supported in windows")
  @unittest.skip("Docking has not been integrated in the latest jaqpotpy version")
  #@pytest.mark.slow
  def test_docker_pose_generator_scores_with_smiles(self):
    from rdkit import Chem
    from rdkit.Chem import AllChem
    import pandas as pd
    """Test that Docker can get scores from pose_generator."""
    # We provide no scoring model so the docker won't score

    smiles_file = "/Users/pantelispanka/guacamol_v1_all.csv"
    df = pd.read_csv(smiles_file)
    for index, smile in enumerate(df[:100]['SMILES']):
        print(index)
        print(smile)
        # pdb_qt = create_hydrated_pdbqt_pdb(self.protein_file)
        # smile = "C1CNCCN(C1)S(=O)(=O)C2=CC=CC3=C2C=CN=C3"
        mol = Chem.MolFromSmiles(smile)
        mol = Chem.AddHs(mol)
        AllChem.EmbedMolecule(mol, randomSeed=0xf00d)
        Chem.MolToMolBlock(mol)
        ligand = "./" + str(index) + "_a_syn_ligand.sdf"
        with Chem.SDWriter(ligand) as w:
            w.write(mol)
        vpg = jp.docking.VinaPoseGenerator(prepare_protein=True)
        docker = jp.docking.Docker(vpg)
        docked_outputs = docker.dock((self.protein_file, ligand),
                                     centroid=(80, 38, 48),
                                     box_dims=(12, 12, 12),
                                     exhaustiveness=12,
                                     num_modes=1,
                                     out_dir="./tmp",
                                     use_pose_generator_scores=True)

        # Check only one output since num_modes==1
        docked_outputs = list(docked_outputs)
        print(docked_outputs)
        print(len(docked_outputs))
        print(len(docked_outputs[0]))
        assert len(docked_outputs) == 1
        assert len(docked_outputs[0]) == 2


  # @unittest.skip("skipping automated test")
  @unittest.skipIf(IS_WINDOWS, "vina is not supported in windows")
  @unittest.skip("Docking has not been integrated in the latest jaqpotpy version")
  #@pytest.mark.slow
  def test_docker_pose_generator_scores_with_pdbqt(self):
    from rdkit import Chem
    from rdkit.Chem import AllChem
    import pandas as pd
    """Test that Docker can get scores from pose_generator."""
    # We provide no scoring model so the docker won't score

    # pdb_qt = create_hydrated_pdbqt_pdb(self.protein_file)
    smile = "C1CNCCN(C1)S(=O)(=O)C2=CC=CC3=C2C=CN=C3"
    mol = Chem.MolFromSmiles(smile)
    mol = Chem.AddHs(mol)
    AllChem.EmbedMolecule(mol, randomSeed=0xf00d)
    Chem.MolToMolBlock(mol)
    ligand = "./fasudil.sdf"
    with Chem.SDWriter(ligand) as w:
        w.write(mol)
    vpg = jp.docking.VinaPoseGenerator(prepare_protein=True)
    docker = jp.docking.Docker(vpg)
    docked_outputs = docker.dock((self.protein_file, ligand),
                                 centroid=(80, 38, 48),
                                 box_dims=(21, 21, 29),
                                 exhaustiveness=18,
                                 num_modes=1,
                                 out_dir="./tmp",
                                 use_pose_generator_scores=True)

    # Check only one output since num_modes==1
    docked_outputs = list(docked_outputs)
    print(docked_outputs)
    print(len(docked_outputs))
    print(len(docked_outputs[0]))
    assert len(docked_outputs) == 1
    assert len(docked_outputs[0]) == 2





  # @unittest.skip("skipping automated test")
  @unittest.skipIf(IS_WINDOWS, "vina is not supported in windows")
  @unittest.skip("Docking has not been integrated in the latest jaqpotpy version")

  #@pytest.mark.slow
  def test_docker_pose_generator_scores(self):
    """Test that Docker can get scores from pose_generator."""
    # We provide no scoring model so the docker won't score
    vpg = jp.docking.VinaPoseGenerator()
    docker = jp.docking.Docker(vpg)
    docked_outputs = docker.dock((self.protein_file, self.ligand_file),
                                 exhaustiveness=4,
                                 num_modes=4,
                                 out_dir="./tmp",
                                 use_pose_generator_scores=True)

    # Check only one output since num_modes==1
    docked_outputs = list(docked_outputs)
    print(docked_outputs)
    print(len(docked_outputs))
    print(len(docked_outputs[0]))
    assert len(docked_outputs) == 4
    assert len(docked_outputs[0]) == 2

  @unittest.skipIf(IS_WINDOWS, "vina is not supported in windows")
  @unittest.skip("skipping automated test")
  #@pytest.mark.slow
  def test_docker_pose_generator_scores_on_pocket(self):
    """Test that Docker can get scores from pose_generator."""
    # We provide no scoring model so the docker won't score
    vpg = jp.docking.VinaPoseGenerator()
    docker = jp.docking.Docker(vpg)
    docked_outputs = docker.dock((self.protein_file, self.ligand_file),
                                 exhaustiveness=40,
                                 num_modes=4,
                                 out_dir="./tmp",
                                 use_pose_generator_scores=True)

    # Check only one output since num_modes==1
    docked_outputs = list(docked_outputs)
    print(docked_outputs)
    assert len(docked_outputs) == 1
    assert len(docked_outputs[0]) == 2

  @unittest.skipIf(IS_WINDOWS, "vina is not supported in windows")
  @unittest.skip("skipping automated test")
  #@pytest.mark.slow
  def test_docker_specified_pocket(self):
    """Test that Docker can dock into spec. pocket."""
    # Let's turn on logging since this test will run for a while
    logging.basicConfig(level=logging.INFO)
    vpg = jp.docking.VinaPoseGenerator()
    docker = jp.docking.Docker(vpg)
    docked_outputs = docker.dock((self.protein_file, self.ligand_file),
                                 centroid=(10, 10, 10),
                                 box_dims=(10, 10, 10),
                                 exhaustiveness=1,
                                 num_modes=1,
                                 out_dir="./tmp")

    # Check returned files exist
    assert len(list(docked_outputs)) == 1

  @unittest.skipIf(IS_WINDOWS, "vina is not supported in windows")
  @unittest.skip("Docking has not been integrated in the latest jaqpotpy version")
  #@pytest.mark.slow
  def test_pocket_docker_dock(self):
    """Test that Docker can find pockets and dock dock."""
    # Let's turn on logging since this test will run for a while
    logging.basicConfig(level=logging.INFO)
    pocket_finder = jp.docking.ConvexHullPocketFinder()
    vpg = jp.docking.VinaPoseGenerator(pocket_finder=pocket_finder)
    docker = jp.docking.Docker(vpg)
    docked_outputs = docker.dock((self.protein_file, self.ligand_file),
                                 exhaustiveness=1,
                                 num_modes=1,
                                 num_pockets=1,
                                 out_dir="./tmp")

    # Check returned files exist
    assert len(list(docked_outputs)) == 1

  #@pytest.mark.slow
  @unittest.skip("Docking has not been integrated in the latest jaqpotpy version")
  def test_scoring_model_and_featurizer(self):
    """Test that scoring model and featurizer are invoked correctly."""

    class DummyFeaturizer(ComplexFeaturizer):

      def featurize(self, complexes, *args, **kwargs):
        return np.zeros((len(complexes), 5))

    class DummyModel(Model):

      def predict(self, dataset, *args, **kwargs):
        return np.zeros(len(dataset))

    class DummyPoseGenerator(PoseGenerator):

      def generate_poses(self, *args, **kwargs):
        return [None]

    featurizer = DummyFeaturizer()
    scoring_model = DummyModel()
    pose_generator = DummyPoseGenerator()
    docker = jp.docking.Docker(pose_generator, featurizer, scoring_model)
    outputs = docker.dock(None)
    assert list(outputs) == [(None, np.array([0.]))]


import subprocess
from jaqpotpy.api.types.models.library import Library


def get_installed_libraries():
    # Run the pip freeze command
    result = subprocess.run(['pip', 'freeze'], stdout=subprocess.PIPE, text=True)
    
    # Split the output into lines
    packages = result.stdout.splitlines()
    
    # Create a list of dictionaries for each package
    packages_list = []
    for package in packages:
        if '==' in package:
            name, version = package.split('==')
            packages_list.append({'name': name, 'version': version})
        elif ' @ ' in package:
            name, _ = package.split(' @ ')
            version = 'Unknown'
            packages_list.append({'name': name, 'version': version})
    
    libraries_list = [Library(library['name'], library['version']) for library in packages_list]
    
    return libraries_list


from torch.utils.data import Dataset
from jaqpotpy.descriptors.graph.graph_featurizer import SmilesGraphFeaturizer
import torch
import pandas as pd
from typing import Iterable, Any, Optional

class SmilesGraphDataset(Dataset):
    """
    A PyTorch Dataset class for handling SMILES strings as graphs.
    This class overrides `__getitem__` and `__len__` (check source code for methods' docstrings).

    Attributes:
        smiles (list): A list of SMILES strings.
        y (list, optional): A list of target values.
        featurizer (SmilesGraphFeaturizer): The object to transform SMILES strings into graph representations.
        precomputed_features (list, optional): A list of precomputed features. If precompute_featurization() is not called, this attribute remains None.
    """

    def __init__(self,
                smiles: list = None,
                y: Optional[list] = None,
                featurizer: Optional[SmilesGraphFeaturizer] = None):
        """
        The SmilesGraphDataset constructor.

        Args:
            smiles (list): A list of SMILES strings. 
            y (list, optional): A list of target values. Default is None.
            featurizer (SmilesGraphFeaturizer, optional): A featurizer object for to create graph representations from SMILES strings.
        
        Example:
        ```
        >>> from jaqpotpy.jaqpotpy_torch.featurizers import SmilesGraphFeaturizer
        >>> from rdkit import Chem
        >>>
        >>> smiles = ['C1=CN=CN1', 'CCCCCCC']
        >>> y = [0, 1]
        >>> featurizer = SmilesGraphFeaturizer()
        >>> featurizer.add_atom_characteristic('symbol', ['C', 'O', 'Na', 'Cl', 'UNK'])
        >>> featurizer.add_atom_characteristic('is_in_ring')
        >>> featurizer.add_bond_characteristic('bond_type', [Chem.rdchem.BondType.SINGLE, Chem.rdchem.BondType.DOUBLE, Chem.rdchem.BondType.TRIPLE])
        >>>
        >>> dataset = SmilesGraphDataset(smiles, y=y, featurizer=featurizer)
        >>> dataset[0]
        Data(x=[5, 6], edge_index=[2, 10], edge_attr=[10, 3], y=0, smiles='C1=CN=CN1')
        ```
        """
        super().__init__()

        self.smiles = smiles
        self.y = y
        # For the moment there is only one Featurizer for Graph
        if featurizer:
            self.featurizer = featurizer
        else:
            self.featurizer = SmilesGraphFeaturizer()
            # Default node, edge features in case of not specifying dataset
            self.featurizer.set_default_config()
        
        self.precomputed_features = None

    # def config_from_other_dataset(self, dataset):
    #     """
    #     Configures the dataset instance from another dataset.

    #     Args:
    #         dataset (SmilesGraphDataset): Another dataset to copy the configuration.

    #     Returns:
    #         SmilesGraphDataset: The configured dataset.
    #     """
    #     self.featurizer = SmilesGraphFeaturizer().config_from_other_featurizer(dataset.featurizer)
    #     return self

    def get_atom_feature_labels(self):
        """
        Returns the atom feature labels.

        Returns:
            list: A list of atom feature labels.
        """
        return self.featurizer.get_atom_feature_labels()

    def get_bond_feature_labels(self):
        """
        Returns the bond feature labels.

        Returns:
            list: A list of bond feature labels.
        """
        return self.featurizer.get_bond_feature_labels()

    def precompute_featurization(self):
        """
        Precomputes the featurization of the dataset before being accessed by __getitem__
        """
        if self.y:
            self.precomputed_features = [self.featurizer(sm, y) for sm, y, in zip(self.smiles, self.y)]
        else:
            self.precomputed_features = [self.featurizer(sm) for sm in self.smiles]

    def get_num_node_features(self):
        """
        Returns the number of node features.

        Returns:
            int: Number of node features.
        """
        return len(self.get_atom_feature_labels())


    def get_num_edge_features(self):
        """
        Returns the number of edge features.

        Returns:
            int: Number of edge features.
        """
        return len(self.get_bond_feature_labels())


    def __getitem__(self, idx):
        """
        Retrieves the featurized graph Data object (and target value) for a given index.

        Args:
            idx (int): Index of the data to retrieve.

        Returns:
            torch_geometric.data.Data: A torch_geometric.data.Data object for this single sample containing:
                - x (torch.Tensor): Node feature matrix with shape [num_nodes, num_node_features].
                - edge_index (LongTensor): Graph connectivity in COO format with shape [2, num_edges].
                - edge_attr (torch.Tensor, optional): Edge feature matrix with shape [num_edges, num_edge_features].
                - y (float): Graph-level ground-truth label.
                - smiles (str): The SMILES string corresponding to the particular sample.
        """
        if self.precomputed_features:
            return self.precomputed_features[idx]
        
        sm = self.smiles[idx]
        y = self.y[idx] if self.y else None

        return self.featurizer(sm, y)

    def __len__(self):
        """
        __len__  functionality is important for the DataLoader to determine batching,
        shuffling and iterating over the dataset. 

        Returns:
            int: Number of samples.
        """
        return len(self.smiles)
    
    def __repr__(self) -> str:
        """
        Official string representation of the Dataset Object
        """
        return self.__class__.__name__




"""
Dataset classes for molecular modelling
"""

from typing import Iterable, Any, Optional
import pandas as pd
from torch.utils.data import Dataset
from jaqpotpy.descriptors.base_classes import MolecularFeaturizer
from jaqpotpy.descriptors.molecular import MolGraphConvFeaturizer
from jaqpotpy.datasets.dataset_base import BaseDataset


class JaqpotpyDataset(BaseDataset):

    """
    A dataset class for general and molecular data, inheriting from BaseDataset. This class is 
    designed to handle datasets that include molecular structures represented by 
    SMILES strings and use molecular featurizers to generate features from these 
    structures.

    Attributes:
        smiles_col (Optional[str]): The column containing SMILES strings.
        smiles (pd.Series): The SMILES strings extracted from the DataFrame.
        featurizer (Optional[MolecularFeaturizer]): The featurizer used to 
                                                    generate molecular features.
        _featurizer_name (Optional[str]): The name of the featurizer.
        x_colls_all (Optional[Iterable[str]]): All feature columns after featurization.
    """

    def __init__(self, df: pd.DataFrame = None, path: Optional[str] = None,
                 y_cols: Iterable[str] = None,
                 x_cols: Optional[Iterable[str]] = None,
                 smiles_cols: Optional[Iterable[str]] = None,
                 featurizer: Optional[MolecularFeaturizer] = None,
                 task:str = None
                 ) -> None:

        if not(isinstance(smiles_cols, str) or 
              (isinstance(smiles_cols, list) and all(isinstance(item, str) for item in smiles_cols)) or
              (isinstance(smiles_cols, list) and len(smiles_cols) == 0) or
              (smiles_cols is None)):
               raise TypeError("smiles_cols should be either a string, an empty list"
                               "a list of strings, or None") 

        if (smiles_cols is not None) and (featurizer is None):
            raise TypeError("Cannot estimate SMILES descriptors without a featurizer"
                            "Please provide a featurizer") 
        
        #Find the length of each provided column name vector and put everything in lists
        if isinstance(smiles_cols, str):
            self.smiles_cols = [smiles_cols]
            self.smiles_cols_len = 1
        elif isinstance(smiles_cols, list) :
            self.smiles_cols = smiles_cols
            self.smiles_cols_len = len(smiles_cols)
        elif smiles_cols is None:
            self.smiles_cols= [] 
            self.smiles_cols_len = 0
        
        if (featurizer is not None) and not (isinstance(featurizer, MolecularFeaturizer)):
            raise TypeError("featurizer should be a MolecularFeaturizer instance ") 

        super().__init__(df=df, path=path, y_cols=y_cols, x_cols=x_cols, task = task)

        self._validate_column_overlap(self.smiles_cols, self.x_cols, self.y_cols)
        self._validate_column_names(self.smiles_cols, "smiles_cols")
        self._validate_column_names(self.x_cols, "x_cols")
        self._validate_column_names(self.y_cols, "y_cols")
        
        self.init_df = self._df
        self.featurizer = featurizer
        self._featurizer_name = None
        self.smiles = None
        self._x_cols_all = None
        self.create()

    @property
    def featurizer_name(self) -> Iterable[Any]:
        return self.featurizer.__name__

    @featurizer_name.setter
    def featurizer_name(self, value):
        self._featurizer_name = value

    @property
    def x_cols_all(self) -> Iterable[str]:
        return self._x_cols_all

    @x_cols_all.setter
    def x_cols_all(self, value):
        self._x_cols_all = value

    def _validate_column_names(self, cols, col_type):
        """
        Validate if the columns specified in cols are present in the DataFrame.
        
        """
        if len(cols) == 0:
            return
        
        missing_cols = [col for col in cols if col not in self._df.columns]

        if missing_cols:
            raise ValueError(f"The following columns in {col_type} are not present in the DataFrame: {missing_cols}")


    def _validate_column_overlap(self, smiles_cols, x_cols, y_cols):
        smiles_set = set(smiles_cols) if smiles_cols else set()
        x_set = set(x_cols) if x_cols else set()
        y_set = set(y_cols) if y_cols else set()

        overlap_smiles_x = smiles_set & x_set
        overlap_smiles_y = smiles_set & y_set
        overlap_x_y = x_set & y_set

        if len(overlap_smiles_x)>0:
            raise ValueError(f"Overlap found between smiles_cols and x_cols: {overlap_smiles_x}")
        if len(overlap_smiles_y)>0:
            raise ValueError(f"Overlap found between smiles_cols and y_cols: {overlap_smiles_y}")
        if len(overlap_x_y)>0:
            raise ValueError(f"Overlap found between x_cols and y_cols: {overlap_x_y}")


    def create(self):

        if  len(self.smiles_cols) == 1:
            # The method featurize_dataframe needs self.smiles to be pd.Series
            self.smiles = self._df[self.smiles_cols[0]]
            descriptors = self.featurizer.featurize_dataframe(self.smiles)
        elif len(self.smiles_cols)>1:
            featurized_dfs = [self.featurizer.featurize_dataframe(self._df[[col]]) for col in self.smiles_cols]
            descriptors = pd.concat(featurized_dfs, axis=1)
        else:
            #Case where no smiles were provided
            self.smiles = []
            descriptors = []
        
        if len(self.x_cols) == 0:
            if len(descriptors)>0:
                self._x = descriptors
                self.x_cols_all = self._x.columns.tolist()
            else:
                raise ValueError("The design matrix X is empty. Please provide either"
                                  "smiles or other descriptors")

        else:
            self._x =  pd.concat([self._df[self.x_cols] ,  pd.DataFrame(descriptors)] , axis=1)
            self.x_cols_all = self._x.columns.tolist()

        if not self.y_cols:
            self._y = None
        else:
            self._y = self._df[self.y_cols]

        self._df = pd.concat([self._x, self._y], axis = 1)
    
    def copy(self):
        """
        Create a copy of the dataset, including a deep copy of the underlying DataFrame
        and all relevant attributes.
        """
        copied_instance = JaqpotpyDataset(
            df= self.init_df,
            path=self.path,
            y_cols=self.y_cols,
            x_cols=self.x_cols,
            smiles_cols=self.smiles_cols,
            featurizer=self.featurizer,
            task=self.task
        )
        return copied_instance

    def __get_X__(self):
        return self._x.copy()
    
    def __get_Y__(self):
        return self._y.copy()
    
    def __get__(self, instance, owner):
        if instance is None:
            return self
        return instance.__dict__[self._df]

    def __getitem__(self, idx):
        selected_x = self._df[self._x].iloc[idx].values
        selected_y = self._df[self._y].iloc[idx].to_numpy()
        return selected_x, selected_y

    def __len__(self):
        return self.df.shape[0]

    def __repr__(self) -> str:
        return (f"{self.__class__.__name__}"
                f"(smiles={True if self.smiles_cols is not None else False}, "
                f"featurizer={self.featurizer_name})")


class TorchGraphDataset(BaseDataset,Dataset):
    """
    Init with smiles and y array
    """
    def __init__(self, smiles=Iterable[str]
                 , y: Iterable[Any] = Iterable[Any], featurizer: MolecularFeaturizer = MolGraphConvFeaturizer(), task=str, streaming: bool = False) -> None:
        super(TorchGraphDataset, self).__init__()
        self._y = y
        self._dataset_name = None
        self._df = []
        self._x_cols_all = None
        self.smiles = smiles
        self._smiles_strings = None
        self.ys = y
        self._task = task
        self.featurizer: MolecularFeaturizer = featurizer
        #self.indices: [] = None
        self.streaming = streaming
        # self.create()

    def create(self):
        if self.streaming is False:
            import torch
            from torch_geometric.data import Data
            descriptors = self.featurizer.featurize(datapoints = self.smiles)
            for i, g in enumerate(descriptors):
                if self._task == 'regression':
                    dato = Data(x=torch.FloatTensor(g.node_features)
                                , edge_index=torch.LongTensor(g.edge_index)
                                , edge_attr= torch.LongTensor(g.edge_features) if g.edge_features is not None else g.edge_features
                                , num_nodes=g.num_nodes, y=torch.Tensor([self.y[i]]))
                    self._df.append(dato)
                elif self._task == 'classification':
                    dato = Data(x=torch.FloatTensor(g.node_features)
                                , edge_index=torch.LongTensor(g.edge_index)
                                , edge_attr=torch.LongTensor(g.edge_features) if g.edge_features is not None else g.edge_features
                                , num_nodes=g.num_nodes, y=torch.LongTensor([self.y[i]]))
                    self._df.append(dato)
                else:
                    raise TypeError("Please set task (classification / regression).")
            self._smiles_strings = self.smiles
            self.X = ['TorchMolGraph']
            self.y = ['Y']
            return self
        else:
            ys = pd.DataFrame(self.ys, columns=['Y'])
            self.df = pd.concat([self.smiles, ys], axis=1)
            return self

    def __getitem__(self, idx):
        if self.streaming is False:
            return self.df[idx]
        else:
            tempdf = []
            import torch
            from torch_geometric.data import Data
            smile = self.smiles[idx]
            descriptors = self.featurizer.featurize(smile)
            for i, g in enumerate(descriptors):
                if self._task == 'regression':
                    dato = Data(x=torch.FloatTensor(g.node_features)
                                , edge_index=torch.LongTensor(g.edge_index)
                                , edge_attr= torch.LongTensor(g.edge_features) if g.edge_features is not None else g.edge_features
                                , num_nodes=g.num_nodes, y=torch.Tensor([self.y[i]]))
                    tempdf.append(dato)
                elif self._task == 'classification':
                    dato = Data(x=torch.FloatTensor(g.node_features)
                                , edge_index=torch.LongTensor(g.edge_index)
                                , edge_attr=torch.LongTensor(g.edge_features) if g.edge_features is not None else g.edge_features
                                , num_nodes=g.num_nodes, y=torch.LongTensor([self.y[i]]))
                    tempdf.append(dato)
                else:
                    raise TypeError("Please set task (classification / regression).")
            self._smiles_strings = self.smiles
            self.X = ['TorchMolGraph']
            self.y = ['Y']
            return tempdf[idx]

    def __len__(self):
        return len(self.df)

    def __repr__(self) -> str:
        # args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
        # args_names = [arg for arg in args_spec.args if arg != 'self']
        # args_info = ''
        # for arg_name in args_names:
        #   value = self.__dict__[arg_name]
        #   # for str
        #   if isinstance(value, str):
        #     value = "'" + value + "'"
          # for list
        return self.__class__.__name__




from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.organization_visibility import OrganizationVisibility
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.model import Model


T = TypeVar("T", bound="Organization")


@_attrs_define
class Organization:
    """
    Attributes:
        name (str):  Example: my-awesome-org.
        contact_email (str):  Example: contact@my-awesome-org.com.
        visibility (OrganizationVisibility):
        id (Union[Unset, int]):
        creator_id (Union[Unset, str]):
        description (Union[Unset, str]):  Example: An awesome organization for managing models..
        user_ids (Union[Unset, List[str]]):
        models (Union[Unset, List['Model']]):
        contact_phone (Union[Unset, str]):  Example: +1234567890.
        website (Union[Unset, str]):  Example: http://www.my-awesome-org.com.
        address (Union[Unset, str]):  Example: 123 Organization St., City, Country.
        can_edit (Union[Unset, bool]): If the current user can edit the organization
        created_at (Union[Unset, str]):
        updated_at (Union[Unset, str]):
    """

    name: str
    contact_email: str
    visibility: OrganizationVisibility
    id: Union[Unset, int] = UNSET
    creator_id: Union[Unset, str] = UNSET
    description: Union[Unset, str] = UNSET
    user_ids: Union[Unset, List[str]] = UNSET
    models: Union[Unset, List["Model"]] = UNSET
    contact_phone: Union[Unset, str] = UNSET
    website: Union[Unset, str] = UNSET
    address: Union[Unset, str] = UNSET
    can_edit: Union[Unset, bool] = UNSET
    created_at: Union[Unset, str] = UNSET
    updated_at: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name

        contact_email = self.contact_email

        visibility = self.visibility.value

        id = self.id

        creator_id = self.creator_id

        description = self.description

        user_ids: Union[Unset, List[str]] = UNSET
        if not isinstance(self.user_ids, Unset):
            user_ids = self.user_ids

        models: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.models, Unset):
            models = []
            for models_item_data in self.models:
                models_item = models_item_data.to_dict()
                models.append(models_item)

        contact_phone = self.contact_phone

        website = self.website

        address = self.address

        can_edit = self.can_edit

        created_at = self.created_at

        updated_at = self.updated_at

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "contactEmail": contact_email,
                "visibility": visibility,
            }
        )
        if id is not UNSET:
            field_dict["id"] = id
        if creator_id is not UNSET:
            field_dict["creatorId"] = creator_id
        if description is not UNSET:
            field_dict["description"] = description
        if user_ids is not UNSET:
            field_dict["userIds"] = user_ids
        if models is not UNSET:
            field_dict["models"] = models
        if contact_phone is not UNSET:
            field_dict["contactPhone"] = contact_phone
        if website is not UNSET:
            field_dict["website"] = website
        if address is not UNSET:
            field_dict["address"] = address
        if can_edit is not UNSET:
            field_dict["canEdit"] = can_edit
        if created_at is not UNSET:
            field_dict["created_at"] = created_at
        if updated_at is not UNSET:
            field_dict["updated_at"] = updated_at

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.model import Model

        d = src_dict.copy()
        name = d.pop("name")

        contact_email = d.pop("contactEmail")

        visibility = OrganizationVisibility(d.pop("visibility"))

        id = d.pop("id", UNSET)

        creator_id = d.pop("creatorId", UNSET)

        description = d.pop("description", UNSET)

        user_ids = cast(List[str], d.pop("userIds", UNSET))

        models = []
        _models = d.pop("models", UNSET)
        for models_item_data in _models or []:
            models_item = Model.from_dict(models_item_data)

            models.append(models_item)

        contact_phone = d.pop("contactPhone", UNSET)

        website = d.pop("website", UNSET)

        address = d.pop("address", UNSET)

        can_edit = d.pop("canEdit", UNSET)

        created_at = d.pop("created_at", UNSET)

        updated_at = d.pop("updated_at", UNSET)

        organization = cls(
            name=name,
            contact_email=contact_email,
            visibility=visibility,
            id=id,
            creator_id=creator_id,
            description=description,
            user_ids=user_ids,
            models=models,
            contact_phone=contact_phone,
            website=website,
            address=address,
            can_edit=can_edit,
            created_at=created_at,
            updated_at=updated_at,
        )

        organization.additional_properties = d
        return organization

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



#from jaqpotpy.datasets import *
#from jaqpotpy.datasets.molecular_datasets import *
from .image_datasets import TorchImageDataset
from .molecular_datasets import JaqpotpyDataset, TorchGraphDataset
from .material_datasets import CompositionDataset, StructureDataset
from .graph_pyg_dataset import SmilesGraphDataset


from typing import Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.model_visibility import ModelVisibility
from ..types import UNSET, Unset

T = TypeVar("T", bound="PartiallyUpdateModelBody")


@_attrs_define
class PartiallyUpdateModelBody:
    """
    Attributes:
        name (str):
        visibility (ModelVisibility):
        description (Union[Unset, str]):
        organization_ids (Union[Unset, List[int]]):
    """

    name: str
    visibility: ModelVisibility
    description: Union[Unset, str] = UNSET
    organization_ids: Union[Unset, List[int]] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name

        visibility = self.visibility.value

        description = self.description

        organization_ids: Union[Unset, List[int]] = UNSET
        if not isinstance(self.organization_ids, Unset):
            organization_ids = self.organization_ids

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "visibility": visibility,
            }
        )
        if description is not UNSET:
            field_dict["description"] = description
        if organization_ids is not UNSET:
            field_dict["organizationIds"] = organization_ids

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        name = d.pop("name")

        visibility = ModelVisibility(d.pop("visibility"))

        description = d.pop("description", UNSET)

        organization_ids = cast(List[int], d.pop("organizationIds", UNSET))

        partially_update_model_body = cls(
            name=name,
            visibility=visibility,
            description=description,
            organization_ids=organization_ids,
        )

        partially_update_model_body.additional_properties = d
        return partially_update_model_body

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.model_type import ModelType
from ..models.model_visibility import ModelVisibility
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.feature import Feature
    from ..models.library import Library
    from ..models.model_meta import ModelMeta
    from ..models.organization import Organization
    from ..models.user import User


T = TypeVar("T", bound="Model")


@_attrs_define
class Model:
    """
    Attributes:
        name (str):  Example: My Model.
        type (ModelType):
        jaqpotpy_version (str):  Example: 1.0.0.
        libraries (List['Library']):
        dependent_features (List['Feature']):
        independent_features (List['Feature']):
        visibility (ModelVisibility):
        actual_model (str): A base64 representation of the actual model.
        id (Union[Unset, int]):
        meta (Union[Unset, ModelMeta]): A JSON object containing meta information.
        description (Union[Unset, str]):  Example: A description of your model.
        organizations (Union[Unset, List['Organization']]):
        pretrained (Union[Unset, bool]):
        creator (Union[Unset, User]):
        can_edit (Union[Unset, bool]): If the current user can edit the model
        tags (Union[Unset, str]):
        created_at (Union[Unset, datetime.datetime]): The date and time when the feature was created. Example:
            2023-01-01T12:00:00Z.
        updated_at (Union[Unset, str]): The date and time when the feature was last updated. Example:
            2023-01-01T12:00:00Z.
    """

    name: str
    type: ModelType
    jaqpotpy_version: str
    libraries: List["Library"]
    dependent_features: List["Feature"]
    independent_features: List["Feature"]
    visibility: ModelVisibility
    actual_model: str
    id: Union[Unset, int] = UNSET
    meta: Union[Unset, "ModelMeta"] = UNSET
    description: Union[Unset, str] = UNSET
    organizations: Union[Unset, List["Organization"]] = UNSET
    pretrained: Union[Unset, bool] = UNSET
    creator: Union[Unset, "User"] = UNSET
    can_edit: Union[Unset, bool] = UNSET
    tags: Union[Unset, str] = UNSET
    created_at: Union[Unset, datetime.datetime] = UNSET
    updated_at: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name

        type = self.type.value

        jaqpotpy_version = self.jaqpotpy_version

        libraries = []
        for libraries_item_data in self.libraries:
            libraries_item = libraries_item_data.to_dict()
            libraries.append(libraries_item)

        dependent_features = []
        for dependent_features_item_data in self.dependent_features:
            dependent_features_item = dependent_features_item_data.to_dict()
            dependent_features.append(dependent_features_item)

        independent_features = []
        for independent_features_item_data in self.independent_features:
            independent_features_item = independent_features_item_data.to_dict()
            independent_features.append(independent_features_item)

        visibility = self.visibility.value

        actual_model = self.actual_model

        id = self.id

        meta: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.meta, Unset):
            meta = self.meta.to_dict()

        description = self.description

        organizations: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.organizations, Unset):
            organizations = []
            for organizations_item_data in self.organizations:
                organizations_item = organizations_item_data.to_dict()
                organizations.append(organizations_item)

        pretrained = self.pretrained

        creator: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.creator, Unset):
            creator = self.creator.to_dict()

        can_edit = self.can_edit

        tags = self.tags

        created_at: Union[Unset, str] = UNSET
        if not isinstance(self.created_at, Unset):
            created_at = self.created_at.isoformat()

        updated_at = self.updated_at

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "type": type,
                "jaqpotpyVersion": jaqpotpy_version,
                "libraries": libraries,
                "dependentFeatures": dependent_features,
                "independentFeatures": independent_features,
                "visibility": visibility,
                "actualModel": actual_model,
            }
        )
        if id is not UNSET:
            field_dict["id"] = id
        if meta is not UNSET:
            field_dict["meta"] = meta
        if description is not UNSET:
            field_dict["description"] = description
        if organizations is not UNSET:
            field_dict["organizations"] = organizations
        if pretrained is not UNSET:
            field_dict["pretrained"] = pretrained
        if creator is not UNSET:
            field_dict["creator"] = creator
        if can_edit is not UNSET:
            field_dict["canEdit"] = can_edit
        if tags is not UNSET:
            field_dict["tags"] = tags
        if created_at is not UNSET:
            field_dict["createdAt"] = created_at
        if updated_at is not UNSET:
            field_dict["updatedAt"] = updated_at

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.feature import Feature
        from ..models.library import Library
        from ..models.model_meta import ModelMeta
        from ..models.organization import Organization
        from ..models.user import User

        d = src_dict.copy()
        name = d.pop("name")

        type = ModelType(d.pop("type"))

        jaqpotpy_version = d.pop("jaqpotpyVersion")

        libraries = []
        _libraries = d.pop("libraries")
        for libraries_item_data in _libraries:
            libraries_item = Library.from_dict(libraries_item_data)

            libraries.append(libraries_item)

        dependent_features = []
        _dependent_features = d.pop("dependentFeatures")
        for dependent_features_item_data in _dependent_features:
            dependent_features_item = Feature.from_dict(dependent_features_item_data)

            dependent_features.append(dependent_features_item)

        independent_features = []
        _independent_features = d.pop("independentFeatures")
        for independent_features_item_data in _independent_features:
            independent_features_item = Feature.from_dict(independent_features_item_data)

            independent_features.append(independent_features_item)

        visibility = ModelVisibility(d.pop("visibility"))

        actual_model = d.pop("actualModel")

        id = d.pop("id", UNSET)

        _meta = d.pop("meta", UNSET)
        meta: Union[Unset, ModelMeta]
        if isinstance(_meta, Unset):
            meta = UNSET
        else:
            meta = ModelMeta.from_dict(_meta)

        description = d.pop("description", UNSET)

        organizations = []
        _organizations = d.pop("organizations", UNSET)
        for organizations_item_data in _organizations or []:
            organizations_item = Organization.from_dict(organizations_item_data)

            organizations.append(organizations_item)

        pretrained = d.pop("pretrained", UNSET)

        _creator = d.pop("creator", UNSET)
        creator: Union[Unset, User]
        if isinstance(_creator, Unset):
            creator = UNSET
        else:
            creator = User.from_dict(_creator)

        can_edit = d.pop("canEdit", UNSET)

        tags = d.pop("tags", UNSET)

        _created_at = d.pop("createdAt", UNSET)
        created_at: Union[Unset, datetime.datetime]
        if isinstance(_created_at, Unset):
            created_at = UNSET
        else:
            created_at = isoparse(_created_at)

        updated_at = d.pop("updatedAt", UNSET)

        model = cls(
            name=name,
            type=type,
            jaqpotpy_version=jaqpotpy_version,
            libraries=libraries,
            dependent_features=dependent_features,
            independent_features=independent_features,
            visibility=visibility,
            actual_model=actual_model,
            id=id,
            meta=meta,
            description=description,
            organizations=organizations,
            pretrained=pretrained,
            creator=creator,
            can_edit=can_edit,
            tags=tags,
            created_at=created_at,
            updated_at=updated_at,
        )

        model.additional_properties = d
        return model

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.organization_invitation_status import OrganizationInvitationStatus
from ..types import UNSET, Unset

T = TypeVar("T", bound="OrganizationInvitation")


@_attrs_define
class OrganizationInvitation:
    """
    Attributes:
        user_email (str): Email address of the invited user
        status (OrganizationInvitationStatus): Status of the invitation
        expiration_date (str): Expiration date of the invitation
        id (Union[Unset, str]): ID of the invitation
        user_id (Union[Unset, str]): The user id associated with that invitation
    """

    user_email: str
    status: OrganizationInvitationStatus
    expiration_date: str
    id: Union[Unset, str] = UNSET
    user_id: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        user_email = self.user_email

        status = self.status.value

        expiration_date = self.expiration_date

        id = self.id

        user_id = self.user_id

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "userEmail": user_email,
                "status": status,
                "expirationDate": expiration_date,
            }
        )
        if id is not UNSET:
            field_dict["id"] = id
        if user_id is not UNSET:
            field_dict["userId"] = user_id

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        user_email = d.pop("userEmail")

        status = OrganizationInvitationStatus(d.pop("status"))

        expiration_date = d.pop("expirationDate")

        id = d.pop("id", UNSET)

        user_id = d.pop("userId", UNSET)

        organization_invitation = cls(
            user_email=user_email,
            status=status,
            expiration_date=expiration_date,
            id=id,
            user_id=user_id,
        )

        organization_invitation.additional_properties = d
        return organization_invitation

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.model import Model


T = TypeVar("T", bound="SearchModelsResponse200")


@_attrs_define
class SearchModelsResponse200:
    """
    Attributes:
        content (Union[Unset, List['Model']]):
        total_elements (Union[Unset, int]):
        total_pages (Union[Unset, int]):
        page_size (Union[Unset, int]):
        page_number (Union[Unset, int]):
    """

    content: Union[Unset, List["Model"]] = UNSET
    total_elements: Union[Unset, int] = UNSET
    total_pages: Union[Unset, int] = UNSET
    page_size: Union[Unset, int] = UNSET
    page_number: Union[Unset, int] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        content: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.content, Unset):
            content = []
            for content_item_data in self.content:
                content_item = content_item_data.to_dict()
                content.append(content_item)

        total_elements = self.total_elements

        total_pages = self.total_pages

        page_size = self.page_size

        page_number = self.page_number

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if content is not UNSET:
            field_dict["content"] = content
        if total_elements is not UNSET:
            field_dict["totalElements"] = total_elements
        if total_pages is not UNSET:
            field_dict["totalPages"] = total_pages
        if page_size is not UNSET:
            field_dict["pageSize"] = page_size
        if page_number is not UNSET:
            field_dict["pageNumber"] = page_number

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.model import Model

        d = src_dict.copy()
        content = []
        _content = d.pop("content", UNSET)
        for content_item_data in _content or []:
            content_item = Model.from_dict(content_item_data)

            content.append(content_item)

        total_elements = d.pop("totalElements", UNSET)

        total_pages = d.pop("totalPages", UNSET)

        page_size = d.pop("pageSize", UNSET)

        page_number = d.pop("pageNumber", UNSET)

        search_models_response_200 = cls(
            content=content,
            total_elements=total_elements,
            total_pages=total_pages,
            page_size=page_size,
            page_number=page_number,
        )

        search_models_response_200.additional_properties = d
        return search_models_response_200

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Dict, Callable


class Preprocess:

    def __init__(self):
        self.classes: Dict[str, Callable] = {}
        self.fitted_classes:  Dict[str, Callable] = {}

        self.classes_y: Dict[str, Callable] = {}
        self.fitted_classes_y:  Dict[str, Callable] = {}

    def register_preprocess_class(self, class_name, class_):
        self.classes[class_name] = class_

    def register_fitted_class(self, class_name, class_):
        self.fitted_classes[class_name] = class_

    def register_preprocess_class_y(self, class_name, class_):
        self.classes_y[class_name] = class_

    def register_fitted_class_y(self, class_name, class_):
        self.fitted_classes_y[class_name] = class_

    def __getitem__(self):
        return self



from typing import Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.feature_type import FeatureType
from ..types import UNSET, Unset

T = TypeVar("T", bound="PartiallyUpdateModelFeatureBody")


@_attrs_define
class PartiallyUpdateModelFeatureBody:
    """
    Attributes:
        name (str): A name for the feature that will appear on top of the form field Example: Updated Feature Name.
        feature_type (FeatureType):  Example: FLOAT.
        description (Union[Unset, str]):  Example: An updated description for this feature.
        possible_values (Union[Unset, List[str]]):
    """

    name: str
    feature_type: FeatureType
    description: Union[Unset, str] = UNSET
    possible_values: Union[Unset, List[str]] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name

        feature_type = self.feature_type.value

        description = self.description

        possible_values: Union[Unset, List[str]] = UNSET
        if not isinstance(self.possible_values, Unset):
            possible_values = self.possible_values

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "featureType": feature_type,
            }
        )
        if description is not UNSET:
            field_dict["description"] = description
        if possible_values is not UNSET:
            field_dict["possibleValues"] = possible_values

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        name = d.pop("name")

        feature_type = FeatureType(d.pop("featureType"))

        description = d.pop("description", UNSET)

        possible_values = cast(List[str], d.pop("possibleValues", UNSET))

        partially_update_model_feature_body = cls(
            name=name,
            feature_type=feature_type,
            description=description,
            possible_values=possible_values,
        )

        partially_update_model_feature_body.additional_properties = d
        return partially_update_model_feature_body

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
